import os,sys,argparse,glob,json,traceback
import numpy as np

def CheckFile(ifile):
    if os.path.isfile(ifile):
        return True
    else:
        print("WARNING: Can not find %s!!!" % ifile)
        return False

def TableOutput(datalist,maxlen=18,digitmax=3,digit=None,left=None,outframe=True,scintific=None,sep=" "):
    'datalist = [[col1,col2..],[col1,col2..]..]'
    context = ''
    ncolumn = len(datalist[0])
    collen = []
    for i in datalist[0]:
        collen.append(len(str(i)))
    for i in range(1,len(datalist)):
        if len(datalist[i]) != ncolumn:
            print("number of column in each line is not same, line %d" % i)
            print(datalist[i])
            return 0
        for j in range(len(datalist[i])):
            if isinstance(datalist[i][j],(float)):
                if scintific != None and scintific[j]:
                    fe = "e"
                else:
                    fe = "f"
                if digit != None :
                    if digit[j] >= 0:
                        lencol = len(("%."+str(digit[j])+fe)%datalist[i][j])
                    else:
                        lencol = len(str(datalist[i][j]))
                else:
                    lencol = len(("%."+str(digitmax)+fe)%datalist[i][j])
            else:
                lencol = len(str(datalist[i][j]))
            if lencol > collen[j] : collen[j] = lencol
    format1 = ''
    totallength = 0
    for i in range(len(collen)):
        if left != None and left[i]:
            lr = "%-"
        else:
            lr = "%"

        length = str(collen[i]) if collen[i] < maxlen else str(maxlen)
        format1 = format1 + lr + length + '.' + length + 's' + sep
        totallength += int(length)

    totallength += len(collen)-1
    if outframe:
        context += totallength*"-" + "\n"

    for i in range(len(datalist)):
        value = []
        for ij,j in enumerate(datalist[i]):
            if isinstance(j,(float,int)):
                if scintific != None and scintific[ij]:
                    fe = "e"
                else:
                    fe = "f"
                if digit != None:
                    if digit[ij] >= 0:
                        j = ("%."+str(digit[ij])+fe) % j
                    else:
                        j = str(j)
                else:
                    j = ("%."+str(digitmax)+fe) % j
            value.append(str(j) )
        context += format1 % tuple(value)
        context += '\n'

    if outframe:
        context += totallength*"-" + "\n"

    return context

def Table2FeishuInteractive(tablelist,webhook,title=None,comment=None,digitmax=3,digit=None,scintific=None):
    """
    translate a table to json used in feishu interactive
    """
    card = {}
    if title:
        #https://open.feishu.cn/document/ukTMukTMukTM/ukTNwUjL5UDM14SO1ATN
        card["header"] = {
            "title": {
                "tag":"plain_text",
                "content": title
            },
            "template":"blue"  #
        }

    #table
    card["elements"] = []
    
    #context
    #https://open.feishu.cn/document/ukTMukTMukTM/ucTNwUjL3UDM14yN1ATN/column-set
    for i,itable in enumerate(tablelist):
        context = []
        for ii,ivalue in enumerate(itable):
            if i > 0 and isinstance(ivalue,(int,float)):
                if digit:
                    idigit = digit[ii]
                else:
                    if isinstance(ivalue,float):
                        idigit = digitmax
                    else:
                        idigit = 0
                if idigit < 0: 
                    idigit = 0
                if scintific:
                    fe = "e" if scintific[ii] else "f"
                else:
                    fe = "f"
                format1 = "%%.%d%s" % (idigit,fe)
                if isinstance(ivalue,bool):
                    value = str(ivalue)
                else:
                    value = format1 % float(ivalue)
            else:
                value = str(ivalue)

            context.append({
                "tag": "column",
                "width": "weighted",
                "weight": 1,
                "elements": [
                {
                  "tag": "markdown",
                  "text_align": "center",
                  "content": value
                }]
            })
        card["elements"].append({
        "tag": "column_set",
        "flex_mode": "stretch",  #stretch, bisect, flow, trisect
        "background_style": "grey" if i == 0 else "default",
        "horizontal_spacing": "default",
        "columns": context})

    if comment:
        card["elements"].append({
        "tag": "markdown",
        "content": comment
        })

    feishu = {
        "msg_type": "interactive",
        "card" :card
    }

    import requests
    response = requests.post(webhook,
        headers={'Content-Type': 'application/json'},
        data=json.dumps(feishu))
    print(response.url,response.status_code,response.text,response.headers)

def GetParamValue(result,param,examplename):
    #result is a dict, whose key is the name of some parameters, and value is the value of the parameter
    #param is the param name, or a formula containing the param name which is identified by '
    #such as: param = "total_energy" or "'total_energy'/'natom'"

    def getvalue(iparam):
        if iparam in result:
            return result.get(iparam,None)
        else:
            if "/" in iparam:
                param_list = iparam.split("/")
                tmp_result = result
                for iiparam in param_list[:-1]:
                    if iiparam not in tmp_result:
                        return None
                    else:
                        tmp_result = tmp_result[iiparam]
                return tmp_result.get(param_list[-1],None)
            else:
                return None


    if "'" in param:
        formula = ""
        readparam = False
        paramname = ""
        tmp_dic = {}
        for i in param:
            if i == "'":
                if not readparam:
                    readparam = True
                    paramname = ""
                else:
                    value = getvalue(paramname) 
                    if value == None:
                        #print(paramname,type(paramname),result)
                        print("%s: value of '%s' is None, set %s to be None" % (examplename,paramname,param))
                        return None
                    
                    if isinstance(value,(int,float)):
                        formula += str(value)
                    elif isinstance(value,list):
                        key = 'np%d'%len(tmp_dic)
                        tmp_dic[key] = np.array(value)
                        formula += "tmp_dic['%s']" % key
                    else:
                        print("%s: value of %s is " % paramname, value,", whose type is %s, is not int or float, set %s to be None" % (examplename,type(value),param))
                        return None
                    readparam = False
            elif readparam:
                paramname += i
            else:
                formula += i
        try:
            #print(formula)
            #print(eval(formula))
            return eval(formula)
        except:
            traceback.print_exc()
            return None       
    else:
        return getvalue(param)

def ProduceExpandDic(allresult,example_idx):
    #construct the expand dict, the key is the key in "outparams_expand"
    #the value is a dict, whose key is "0", "1", .. "ntype"
    # and whose value is the value defined in outparams.
    expand_dic = {}
    outparams = allresult.get("outparams",[])
    outparams_paramname = [i[0] for i in outparams]
    example_name = allresult.get("example_name",[])
    results = allresult.get("result",[[]])
    type_idx = allresult.get("type_idx",[])
    #print(outparams_paramname)
    for key in allresult.get("outparams_expand",{}):
        if key not in outparams_paramname:
            print("key '%s' defined in 'outparams_expand' is not defined in 'outparams', skip it" % key)
            continue
        expand_dic[key] = {}
        for i in range(len(results)):
            result = GetParamValue(results[i][example_idx],
                                   outparams[outparams_paramname.index(key)][1][type_idx[i]],
                                   example_name[example_idx])
            expand_dic[key]['%d'%i] = result
    #print(expand_dic)
    return expand_dic

def OutParam(allresult,split_example="----"):
    '''
    allresult = {
        "result": [[type1_result1,type1_result2,...],[type2_result1,type2_result2,..],...],
        "type_name": [str],
        "example_name": ["example1","example2",...],
        "outparams": [[param1_name_in_table, [param1_name_in_typeAresult,param_name1_in_typeBresult,...], significant_digit_number, False],
                      [param2_name_in_table, [param2_name_in_typeAresult,param_name2_in_typeBresult,...], significant_digit_number, False],
                      ...],
        "type_idx": [],
        "outparams_expand": {str:[]},
        "outparams_comment": ""
    }
    
    type_name: the name of each job type.
    type_result: is a dictionary, whose key is param name.
    type_idx: ntype elemens. Each element is the index outparams[][1], which indicate the param name in this job type.
    outparams_expand: a dictionary. The key is outparams[][0], and value is a list of each job type. You can define the calculation between different job types.
                        Such as: to calculate the enery difference between job type 0 and 1, and show in line of type0, you can write as {"energy": ["'0' - '1'",""]}
    
    '''
    outtable = [['example',"jobType"]]
    param = allresult.get("outparams")
    for i in param:
        if i[-1] != None: outtable[0].append(i[0])
           
    ncol = len(outtable[0]) 
    outtable.append(ncol * ["----"])
    
    results = allresult.get("result",[[]])
    example_name = allresult.get("example_name",[])
    type_idx = allresult.get("type_idx",[])
    allparam_value = [[] for i in range(len(results))] #[[type1_example1_dict, type1_example2_dict],[type2_...]]
    for i in range(len(results[0])):  #i for example
        expand_dic = ProduceExpandDic(allresult,i)
        for j in range(len(results)):  # j for job type
            iexample = [example_name[i]] if j==0 else [" "]   #add example name
            iexample.append(allresult["type_name"][j])
            allparam_value[j].append({})
            for iparam in param:
                if iparam[0] not in expand_dic:
                    value = GetParamValue(results[j][i],iparam[1][type_idx[j]],example_name[i])
                else:
                    expand_formula = allresult['outparams_expand'][iparam[0]]
                    if j < len(expand_formula) and expand_formula[j].strip() != "":
                        value = GetParamValue(expand_dic[iparam[0]],expand_formula[j],example_name[i]+"/"+iparam[0])
                    else:
                        value = " "
                if iparam[-1] != None:
                    iexample.append(value) 
                #print(allparam_value[j][-1])
                allparam_value[j][-1][iparam[0]] = value if value != " " else None 
            outtable.append(iexample)
        if split_example != None: outtable.append(ncol * [split_example])
    if split_example != None: outtable = outtable[:-1]
    
    
    digit = [-1,-1] + [i[2] for i in param]
    left = [True,True]
    for i in param:
        if len(i) == 4:
            left.append(i[3])
        else:
            left.append(False)
            
    cc = '\nSome key results of ' + ", ".join(allresult.get("type_name")) + "\n"
    comment = allresult.get("outparams_comment",[])
    if len(comment) > 0:
        cc += "\n".join(comment) + "\n"
    cc += TableOutput(outtable,maxlen=18,digit=digit,left=left)
    if allresult.get("webhook"):
        Table2FeishuInteractive(outtable,webhook=allresult.get("webhook"),title="abacustest",digit=digit)
    return cc,allparam_value

class MetricsMethod:
    @classmethod
    def allmethod(cls):
        a = cls()
        method = {}
        for k,v in a.method_list().items():
            method[k] = v[0]
        return method
    
    @classmethod
    def allmethod_str(cls):
        a = cls()
        method = ''
        for k,v in a.method_list().items():
            method += "%10s: %s\n" % (k,v[1])
        return method
    
    def method_list(self):
        return {"GM":[self.GM,"calculate the geometric mean value"],
                "iGM":[self.iGM, "calculate the geometric mean of the inverse of value"],
                "MEAN":[self.MEAN,"mean of all values"],
                "TrueRatio":[self.TrueRatio,"the ratio of True"]
                }
    def MEAN(self,valuelist):
        return np.array(valuelist).mean()
        
    def CalGM(self,valuelist,func=lambda x:x):
        n = len(valuelist)
        gm = 1
        for i in valuelist:
            gm *= func(i)
        return (gm)**(1/n)
    
    def GM(self,valuelist):
        return self.CalGM(valuelist)
    
    def iGM(self,valuelist):
        return self.CalGM(valuelist,lambda x: 1/x)
    
    def TrueRatio(self,valuelist):
        nTrue = 0
        for i in valuelist:
            if i == True:
                nTrue += 1
        
        return float(nTrue)/float(len(valuelist))
        
def OutMetrics(allresults,allparam_value):
    type_name = allresults.get("type_name")
    example_name = allresults.get("example_name",[])
    results = allresults.get("result",[[]])
    
    outtable = [["metrics"] + type_name + ['example_number'],(len(type_name)+2)*["----"]]
    metrics = allresults.get("metrics",[])

    method_list = MetricsMethod.allmethod()
    hasnotsupportmethod = False
    notsupportmethod = []
    comment = []
    allmetric_value = {}
    for metric in metrics:
        name = metric.get("name")
        paramname = metric.get("param_name") 
        method = metric.get("method")
        condition = metric.get("condition","").strip() #if condition is not meet, value will be set to None
        doclean = metric.get("doclean",False)  #if remove the value of None
        normalization = metric.get("normalization",True)
        if "comment" in metric:
            icomment = metric["comment"].strip()
            if icomment != "":
                comment.append("%s:%s\n" % (name,icomment))
        #if paramname not in outparams_paramname:
        #    print("key '%s' defined in 'metrics/param_name' is not defined in 'outparams', skip it" % paramname) 
        #    continue
        if  method not in method_list:
            if method not in notsupportmethod: notsupportmethod.append(method)
            hasnotsupportmethod = True
            continue
        tmp_result = [[] for i in range(len(type_name))]
        for i in range(len(results[0])):
            tmp = []
            hasnone = False
            for j in range(len(results)):
                tmp_name = type_name[j] + "/" + example_name[i]
                if condition == "":
                    value = GetParamValue(allparam_value[j][i],paramname,tmp_name)
                else:
                    conditionvalue = GetParamValue(allparam_value[j][i],condition,tmp_name)
                    if conditionvalue:
                        value = GetParamValue(allparam_value[j][i],paramname,tmp_name)
                    else:
                        value = None
                if value == None:
                    hasnone = True
                tmp.append(value)
            if doclean and hasnone:
                continue
            else:
                for j,itmp in enumerate(tmp): 
                    if itmp!=None: tmp_result[j].append(itmp)
        if doclean and len(tmp_result[0]) == 0:
            print("%s: values of all examples are None for some job type, skipt it!" % name)
            continue
        
        nresult = [str(len(i)) for i in tmp_result]
        nexample = nresult[0] if doclean else "/".join(nresult)
               
        outtable.append([name])
        metric_value = []
        for i in range(len(tmp_result)):
            try:
                metric_value.append(method_list[method](tmp_result[i]))
            except:
                traceback.print_exc()
                metric_value.append(None)

            if normalization:
                try:
                    outtable[-1].append(metric_value[i]/metric_value[0]) #do normalization by divide by the first value
                except:
                    traceback.print_exc()
                    outtable[-1].append(None)
            else:
                outtable[-1].append(metric_value[i])        

        if normalization and metric_value[0] != None:
            comment += "%-20s is normalized, value of %s is %.3e.\n" % (name,type_name[0],metric_value[0])

        outtable[-1] += [nexample]
        allmetric_value[name] = metric_value
        
    if hasnotsupportmethod:
        print("Method %s are not supportted now.\nSupported methods are:\n%s" % (" ".join(notsupportmethod),MetricsMethod.allmethod_str()))

    digit = [-1] + len(type_name)*[3] + [0]
    left = [True] + (len(type_name)+1)*[False]
    scintific = (len(type_name)+2)*[False]
    
    cc = "\nSome key metrics\n"
    cc += TableOutput(outtable,maxlen=50,digit=digit,left=left,scintific=scintific)
    if allresults.get("webhook"):
        Table2FeishuInteractive(outtable[0:1] + outtable[2:],webhook=allresults.get("webhook"),title="abacustest",digit=digit)
    cc += "".join(comment)
    #cc += "Notice: the exmaples with value of None for some job type will be excluded.\n"
    return cc,allmetric_value    

def GetAllResults(result_setting):
    #allresults = {"type_name":[types],"result_file": [result1,result2,result3,....] }
    #result is a two dimension list, 1st dimension is for result of each example in each type,
    #second dimension is types
    #len("example_name") = len("result"[i]) = len("result"[j])
    #len("type_name") = len("result")
    #result123 = [example1_result, example2_result, example3_result]

    resultf = result_setting["result_file"]
    type_name = result_setting.get("type_name",["jobtype"])
    
    #deal with type_name
    if isinstance(type_name,str):
        if len(resultf) > 1:
            type_name = [type_name+str(i) for i in range(len(resultf))]
        else:
            type_name = [type_name]
    elif isinstance(type_name,list) and len(type_name) == 1 and len(resultf) > 1:
            type_name = [type_name[0]+str(i) for i in range(len(resultf))]
    if not isinstance(type_name,list):
        print("type_name should be a list of type names")
        return False
    if (len(resultf) != len(type_name)):     
        print("number of result_fiel is not equal to type_name")
        return False
    
    #deal with example_name_idx
    example_name_idx = result_setting.get("example_name_idx",-1) 
    if isinstance(example_name_idx,int):
        example_name_idx = len(type_name) * [example_name_idx]
    elif isinstance(example_name_idx,list) and len(example_name_idx) == 1 and len(type_name) > 1:
        example_name_idx = len(type_name) * example_name_idx
    if not isinstance(example_name_idx,list):
        print("example_name_idx should be a list of int")
        return False        
    if (len(example_name_idx) != len(type_name)):
        print("number of example_name_idx is not equal to type_name")
        return False
    
    #deal with type_idx
    type_idx = result_setting.get("type_idx",[0 for i in range(len(type_name))])
    if isinstance(type_idx,int):
        type_idx = len(type_name) * [type_idx]
    elif isinstance(type_idx,list) and len(type_idx) == 1 and len(type_name) > 1:
        type_idx = len(type_name) * type_idx
    if not isinstance(type_idx,list):
        print("type_idx should be a list of int")
        return False 
    if (len(type_idx) != len(type_name)):
        print("number of type_idx is not equal to type_name")
        return False
    
    #deal with results
    result = []
    example_name = []
    for i,iresultf in enumerate(resultf):
        results = json.load(open(iresultf))
        result.append([])
        if i==0:
            for k,v in results.items():
                example_name.append(k.split("/")[example_name_idx[i]])
                result[0].append(v)
        else:
            tmp_dict = {}
            for k,v in results.items():
                tmp_dict[k.split("/")[example_name_idx[i]]] = v
            for iexample in example_name:
                result[-1].append(tmp_dict.get(iexample,{}))         
    
    #print("example number: %d" % (len(example_name)))
    #print("type number: %d" % (len(type_name)))
    #print("type idx: %d" % (len(type_idx)))

    hasfalse = False
    if len(result) != len(type_name):
        print("type number is %d, but the type of result is %d" % (len(type_name),len(result)))
        hasfalse = True
        
    if len(type_idx) != len(type_name):
        print("type idx is %d, but the type number is %d" % (len(type_idx),len(type_name)))
        hasfalse = True
        
    for i,ir in enumerate(result):
        if len(ir) != len(example_name):
            print("type %d has %d results, which is not equal to example number %d" % (i+1,len(ir),len(example_name)))
            hasfalse = True

    if hasfalse:
        return False

    outparams = result_setting.get("outparams",[])
    metrics = result_setting.get("metrics",[])
    outparams_expand = result_setting.get("outparams_expand",{})
    outparams_comment = result_setting.get("outparams_comment",{})
    plot = result_setting.get("plot",[])
    webhook = result_setting.get("webhook",None)
    
    return {"example_name": example_name, 
            "type_name": type_name, 
            "result": result,
            "type_idx": type_idx, 
            "outparams": outparams,
            "outparams_expand":outparams_expand,
            "outparams_comment":outparams_comment,
            "metrics": metrics,
            "plot":plot,
            "webhook":webhook}
    
def check_file(ifile):
    # check if ifile exist,
    # if exist, then modify the file name to add a number at the end until the file does not exist
    # return the new file name
    ifile = os.path.abspath(ifile)
    if os.path.isfile(ifile):
        ifile_prefix = os.path.splitext(ifile)[0]
        ifile_suffix = os.path.splitext(ifile)[1]
        i = 1
        while os.path.isfile(ifile):
            ifile = ifile_prefix + "_%d" % i + ifile_suffix
            i += 1
    return ifile

def pandas_out(allresult,savefile = None, report_sample_max = 2,print_result=True, print_list_seperate = False):
    """
    allresult = {sample1: {key1:value,key2:value},
                 sample2: {key1:value,key2:value}}
    If value is a list, will print out separately.
    """
    import pandas as pd
    pd.set_option('display.max_columns', None)
    pd.set_option('display.max_rows', None)
    pd.set_option('display.width', 200)

    normal_result = {}
    list_result = []
    allsamples = [i for i in allresult.keys()]
    allkeys = []
    for v in allresult.values():
        for iv in v.keys():
            if iv not in allkeys:
                allkeys.append(iv)
        
    allkeys_seperate = []
    savefile_prefix = "" if not savefile else os.path.splitext(savefile)[0]
    savefile_names = []
    for ikey in allkeys:
        allkeys_seperate.append(False)
        if not print_list_seperate: # if not print list separately, then print out the list together
            continue
        for i in allsamples:
            if isinstance(allresult[i].get(ikey,None),dict):
                allkeys_seperate[-1] = True
                break
            elif isinstance(allresult[i].get(ikey,None),list):
                #if len(allresult[i][ikey]) > 0 and isinstance(allresult[i][ikey][0],(list,dict)):
                    allkeys_seperate[-1] = True
                    break
    nsample = 0    
    for isample in allsamples:
        nsample += 1
        normal_result[isample] = {}
        for i,ikey in enumerate(allkeys):
            sfname = savefile_prefix +"_" + isample.replace("/","_")+"_"+ikey.replace("/","_")
            if allkeys_seperate[i]:
                if isinstance(allresult[isample].get(ikey,None),(list,dict)):
                    try:
                        list_result.append("%s\t%s:\n" % (isample,ikey) + str(pd.DataFrame.from_dict(allresult[isample][ikey])))
                        if savefile:
                            fname_final = check_file(sfname+".csv")
                            pd.DataFrame.from_dict(allresult[isample][ikey]).to_csv(fname_final)
                            if nsample <= report_sample_max:
                                savefile_names.append(fname_final)
                            
                    except:
                        list_result.append("%s\t%s:\n" % (isample,ikey) + str(allresult[isample][ikey]))
                        if savefile:
                            fname_final = check_file(sfname+".txt")
                            with open(fname_final,'w') as f:
                                f.write(str(allresult[isample][ikey]))
                                if nsample <= report_sample_max:
                                    savefile_names.append(fname_final)
                            
                else:
                    if allresult[isample].get(ikey,None) != None:
                        list_result.append("%s\t%s:\n" % (isample,ikey) + str(allresult[isample].get(ikey,None)))
                        if savefile:
                            fname_final = check_file(sfname+".txt")
                            with open(fname_final,'w') as f:
                                f.write(str(allresult[isample][ikey]))
                                if nsample <= report_sample_max:
                                    savefile_names.append(fname_final)
            else:
                normal_result[isample][ikey] = allresult[isample].get(ikey,None)
                
    if False not in allkeys_seperate:
        normal_result = ""
    else:
        pddata = pd.DataFrame.from_dict(normal_result,orient='index')
        if savefile:
            savefile_name_final = check_file(savefile)
            pddata.to_csv(savefile_name_final, index_label="example")
            savefile_names.insert(0,savefile_name_final)
        normal_result = str(pddata)
        
    if True not in allkeys_seperate:
        list_result = ""
    else:
        list_result = "\n\n".join(list_result)
    
    if print_result:
        print("%s\n\n%s" % (normal_result,list_result))      
    return normal_result,list_result,savefile_names

def OutResultArgs(parser):  
    parser.description = "This script is used to output the summary of results"
    parser.add_argument('-r', '--result', type=str, help='the result file from collectdata, should be .json type',action="extend",nargs="*")
    parser.add_argument('-p', '--param', type=str, help='the parameter file, should be .json type')
    parser.add_argument('-m', '--metrics', default=["ALL"], help='The metrics that needed to be shown. Use ALL to show all metrics, default is ALL.', action="extend",nargs="*")
    parser.add_argument('-o', '--output', type=str, help='output the selected metrics to a json file')
    parser.add_argument('--csv', type=str, help='if output the selected metrics to csv file')
    parser.add_argument("--list_sep", default=0, const=1, nargs='?', type=int, help="if print the list separately, default is 0")
    
    return parser

def outresult(param):
    if param.result != None:
        allresult_files = param.result #if len(param.result) == 1 else param.result[1:]
        metric_name = param.metrics if len(param.metrics) == 1 else param.metrics[1:]
        allresult = {}
        allfiles = []
        for ifile in allresult_files:
            for iif in glob.glob(ifile):
                allfiles.append(iif) 
        for ifile in allfiles:
            result = json.load(open(ifile))
            for k,v in result.items():
                key = "%s:%s" % (ifile,k) if len(allfiles) > 1 else k
                allresult[key] = v
        if "ALL" not in metric_name:
            need_metric = []
            for i in metric_name:
                if i not in need_metric:
                    need_metric.append(i)
            new_result = {}
            for ik,iv in allresult.items():
                new_result[ik] = {}
                for i in need_metric:
                    new_result[ik][i] = iv.get(i,None)
            allresult = new_result
        if param.output != None:
            json.dump(allresult,open(param.output,'w'),indent=4)
        pandas_out(allresult,param.csv,print_list_seperate=param.list_sep)
    
    if param.param!= None:
        if not CheckFile(param.param):
            sys.exit(1)
        
        report_part = json.load(open(param.param))
        if "allresults" in report_part:
            report_part = report_part["allresults"]
        #elif "report" in report_part:
        #    report_part = report_part["report"]
        allresults = GetAllResults(report_part)
        if len(allresults['type_name']) == 1: split_example = None
        else: split_example = "----"
        cc_outparam,allparam_value = OutParam(allresults,split_example=split_example)
        cc_outmetrics,allmetric_value = OutMetrics(allresults,allparam_value)
        
        output_value = {}
        output_value['type_name'] = allresults['type_name']
        output_value['example_name'] =  allresults['example_name']
        output_value['param_value'] = allparam_value
        output_value['metrics_value'] = allmetric_value
        
        json.dump(allmetric_value,open("superMetrics.json","w"),indent=4)
        if param.output != None:
            json.dump(output_value,open(param.output,'w'),indent=4)
        if allmetric_value:
            print(cc_outmetrics)
        print(cc_outparam)

def main():
    parser = argparse.ArgumentParser()
    param = OutResultArgs(parser).parse_args()
    outresult(param)
    
if __name__ == "__main__":
    main()


from dp.launching.typing.basic import BaseModel, Int, String, Float,List,Optional,Union,Dict
from dp.launching.typing import InputFilePath, OutputDirectory
from dp.launching.typing import (
    Field,
    DataSet,
)
from dp.launching.report import Report,AutoReportElement,ReportSection

from . import comm_class,comm_func,comm_class_exampleSource,comm_pmetrics
import json,traceback,os

import dp.launching.typing.addon.ui as ui
from dp.launching.typing.addon.sysmbol import Equal,NotEqual,Exists,NotExists

class SelfDefine(BaseModel):
    setting_upload: InputFilePath = Field(default=None,
                                          title="Upload setting file",
                                          st_kwargs_type=["json"],
                                          description="Use setting file by uploading. (priority: setting string > setting upload > setting dataset)",
                                          description_type="markdown")

    setting_dataset: DataSet = Field(title="setting dataset",
                                     default=None,
                                     description="Use setting file from dataset. (priority: setting string > setting upload > setting dataset)",)

    setting_string: String = Field(
        title="setting string",
        default="",
        description="Use setting by enter string. (priority: setting string > setting upload > setting dataset)",)
  

class SelfDefineModel(comm_class.TrackingSet,
                      SelfDefine,
                      comm_class_exampleSource.ExampleSet,
                      comm_class_exampleSource.ExampleSourceSet,
                      comm_class.ConfigSet,
                      comm_class.OutputSet,
                      BaseModel):
    ...

class SelfDefineDatasetsModel(comm_class.TrackingSet,
                      SelfDefine,
                      comm_class_exampleSource.ExampleSet,
                      comm_class_exampleSource.DatasetSet,
                      comm_class.ConfigSet,
                      comm_class.OutputSet,
                      BaseModel):
    ...
    
def SelfDefineModelRunner(opts):
    logs = comm_class.myLog()  
    paths = comm_func.create_path(str(opts.IO_output_path))
    output_path = paths["output_path"]
    work_path = paths["work_path"]
    download_path = paths["download_path"]

    logs.iprint("read source setting ...")
    datas = comm_class_exampleSource.read_source(opts,work_path,download_path,logs.iprint)
    if datas == None or not datas.get("all_files"):
        logs.iprint("Error: download examples or rundft_extrafiles or postdft_extrafiles failed!")
        return 1

    # parse inputs
    # setting inputs is prefered
    # and then the uploaded file
    # and then the file in dataset or examples
    if opts.setting_string.strip() != "":
        try:
            setting = json.loads(opts.setting_string)
        except:
            traceback.print_exc()
            return 1
    elif opts.setting_upload != None:
        try:
            setting = json.load(open(opts.setting_upload.get_path()))
        except:
            traceback.print_exc()
            return 1
    elif opts.setting_dataset:
        try:
            setting = json.load(open(opts.setting_dataset.get_full_path()))
        except:
            traceback.print_exc()
            return 1
    else:
        print("Please supply the setting information")
        return 1

    #read setting
    allparams = {"config": comm_func.read_config(opts)}
    for k,v in setting.items():
        if k == "config":
            for ik,iv in v.items():
                allparams["config"][ik] = iv
        #elif k in ["ABBREVIATION","save_path","run_dft","post_dft","report","dataset_info","upload_datahub","upload_tracking"]:
        else:
            allparams[k] = v
            
    tracking_set = comm_class.TrackingSet.parse_obj(opts)
    if tracking_set:
        if "aim_access_token" not in allparams["config"]:
            allparams["config"]["aim_access_token"] = str(tracking_set.get("token"))
        if "post_dft" not in allparams:
            allparams["post_dft"] = {}
        if "upload_tracking" not in allparams["post_dft"]:
            allparams["post_dft"]["upload_tracking"] = {}
        if "name" not in allparams["post_dft"]["upload_tracking"]:
            allparams["post_dft"]["upload_tracking"]["name"] = tracking_set.get("name")
        if "experiment" not in allparams["post_dft"]["upload_tracking"]:
            allparams["post_dft"]["upload_tracking"]["experiment"] = tracking_set.get("experiment")
        if "tags" not in allparams["post_dft"]["upload_tracking"]:
            allparams["post_dft"]["upload_tracking"]["tags"] = tracking_set.get("tags")
            
    #execut
    stdout,stderr = comm_func.exec_abacustest(allparams,work_path)
    logs.iprint(f"{stdout}\n{stderr}\nrun abacustest over!\n")
    reports = comm_pmetrics.produce_metrics_superMetrics_reports(allparams,work_path,output_path)
    
    logfname = "output.log"
    logs.write(os.path.join(str(opts.IO_output_path),logfname))
    log_section = ReportSection(title="",
                              elements=[AutoReportElement(title='', path=logfname, description="")])
    reports.append(log_section)

    if reports:
        report = Report(title="abacus test report",
                        sections=reports,
                        description="a report of abacustest")
        report.save(output_path)
        
    #move results to output_path
    comm_func.move_results_to_output(work_path,output_path,allparams.get("save_path","results"))
    comm_func.pack_results(output_path,allparams.get("save_path","results"))
    
    return 0



import os,sys,argparse,glob,json,traceback,re,shutil
import numpy as np
from typing import Dict, List, Optional, Union
import copy
from pathlib import Path
from abacustest import constant
from abacustest.myflow import comm
from abacustest.lib_prepare import abacus as MyAbacus
from abacustest.lib_prepare import abacus2qe as Aba2Qe
from abacustest.lib_prepare import abacus2vasp as Aba2Vasp
from abacustest.lib_prepare import abacus2cp2k as Aba2Cp2k
from abacustest.lib_prepare.comm import translate_strus, collect_pp

def Direct2Cartesian(coord:List[List[float]],cell:List[List[float]]):
    return np.array(coord).dot(np.array(cell)).tolist()

def Cartesian2Direct(coord:List[List[float]],cell:List[List[float]]):
    return np.array(coord).dot(np.linalg.inv(np.array(cell))).tolist()
 
class PrepareAbacus:
    def __init__(self,
                 save_path: str = Path("."),
                 example_template: str = None,
                 input_template: str = None,
                 kpt_template: str = None,
                 stru_template: str = None,
                 mix_input: Dict[str,any] = {},
                 mix_kpt: List[Union[int,List[int]]] = [],
                 mix_stru: List[str]=[],
                 pert_stru: Dict[str,any] = {},
                 pp_dict: Dict[str,str]= {},
                 orb_dict: Dict[str,str]= {},
                 paw_dict: Dict[str,str]= {},
                 pp_path:str = None,
                 orb_path:str = None,
                 dpks_descriptor:str=None,
                 extra_files: List[str] = [],
                 bak_file = True,
                 no_link = False,
                 link_example_template_extra_files=True,):
        """To prepare the inputs of abacus

        Parameters
        ----------
        save_path : str, optional
            the path to store inputs, by default Path(".")

        example_template : str, optional
            folder name of an abacus inputs template, by default None.
            If specify input/kpt/stru_template, the related file in exmaple_template will
            be ingored. The pp/orb defined in example_template/STRU will also added to 
            pp/orb_dict.

        input_template : str, optional
            file name of INPUT template, by default None

        kpt_template : str, optional
            file name of KPT template, by default None

        stru_template : str, optional
            file name of STRU template, by default None

        mix_input : Dict[str,any], optional
            mix setting of input parameters, by default {}. 
            Such as:{"ecutwfc":[50, 60, 70], 
                     "mixing_beta":[0.3, 0.4, 0.5]}, 
            then will prepare 3*3 = 9 inputs.

        mix_kpt : List, optional
            mixing setting of k point list, by default []. If this parameter
            is defined, then kpt_template will be invalide.
            Such as: [[1,1,1,0,0,0],
                      [2,2,2,0,0,0]], 
            then will prepare 2 KPT files.
            The kpoint list can be a list of 6 values (3 int number for K point
            and 3 float values for the shift value) like: [1,1,1,0,0,0], 
            or 3 values like: [1,1,1] which means [1,1,1,0,0,0], or one int number
            like: 3 which means [3,3,3,0,0,0].

        mix_stru : List, optional
            a list of STRU files, by default []. If this parameter is defined
            then stru_template will be invalide.
            Such as: ["a/STRU","b/STRU"]
            
        pert_stru : Dict, optional
            perturbation setting of structure, by default {}.
            
        pp_dict : Dict, optional
            a dictionary specify the pseudopotential files, by default {}

        orb_dict : Dict, optional
            a dictionary specify the orbital files, by default {}
        
        paw_dict: Dict, optional
            a dictionary specify the paw files, by default {}

        pp_path : str, optional
            the path of your pseudopotential lib, by default None. 
            The file name shuold be started with the element name and followed by "_".
            Such as: Ag_ONCV_PBE-1.0.upf.

        orb_path : str, optional
            the path of your orbital lib, by default None. 
            The file name shuold be started with the element name and followed by "_".
            Such as: Ag_gga_7au_100Ry_4s2p2d1f.orb.

        dpks_descriptor : str, optional
            file name of deepks descriptor, by default None

        extra_files : List, optional
            a list of somes extra files that will be putted in each input, by default []   
            
        bak_file: bool, optional
            when save path is exist, if bak_file is True, then bak the old files,
            
        no_link: bool, optional
            if True, then will not link the files in example_template to save_path,
            
        link_example_template_extra_files: bool, optional
            if True, then will link the extra files in example_template to save_path,
        """        
        self.save_path = save_path
        self.example_template = example_template
        self.input_template = input_template
        self.kpt_template = kpt_template
        self.stru_template = stru_template
        self.mix_input = mix_input
        self.pp_dict = pp_dict
        self.orb_dict = orb_dict
        self.paw_dict = paw_dict
        self.extra_files = self.CheckExtrafile(extra_files)
        self.mix_kpt = mix_kpt
        self.mix_stru = mix_stru
        self.pert_stru = pert_stru
        self.dpks_descriptor = dpks_descriptor if dpks_descriptor else None
        
        self.pp_path = None if not pp_path else pp_path.strip()
        self.orb_path = None if not orb_path else orb_path.strip()
        self.CollectPP()
        self.CollectOrb()
        
        self.input_list,self.input_mix_param = self.Construct_input_list()
        self.kpt_list = self.Construct_kpt_list()
        self.stru_list = self.Construct_stru_list()
        
        self.bak_file = bak_file
        self.no_link = no_link
        
        # when read the template file, will check if the template file folder is same with save_path
        # if same and final structures only one, then will not bak the template folder
        self.template_is_save_path = self.CheckIfTemplateIsSavePath()
        self.example_template_extra_files = self.CheckExampleTemplateExtraFiles(link_example_template_extra_files)
    
    def CheckExampleTemplateExtraFiles(self,link_example_template_extra_files):
        # check the extra files in example_template except for INPUT, KPT, STRU
        if (not link_example_template_extra_files) or (not self.example_template):
            return []
        if not os.path.isdir(self.example_template):
            return []
        allfiles = os.listdir(self.example_template)
        extrafiles = []
        for ifile in allfiles:
            if ifile in ["INPUT","STRU","KPT"]:
                continue
            if os.path.isfile(os.path.join(self.example_template,ifile)):
                extrafiles.append(os.path.abspath(os.path.join(self.example_template,ifile)))
        return extrafiles
        
    def CheckIfTemplateIsSavePath(self):
        if not os.path.exists(self.save_path):
            return False
        CheckedPath = []
        if self.input_template:
            CheckedPath.append(Path(os.path.split(self.input_template)[0]))
        if self.kpt_template:
            CheckedPath.append(Path(os.path.split(self.kpt_template)[0]))
        if self.stru_template:
            CheckedPath.append(Path(os.path.split(self.stru_template)[0]))
        if self.example_template:
            CheckedPath.append(Path(self.example_template))
        for ipath in CheckedPath:
            if Path(self.save_path).samefile(ipath):
                return True
        return False

    def CheckExtrafile(self,extrafiles):
        if not extrafiles:
            return []
        filelist = []
        for ifile in extrafiles:
            if os.path.isfile(ifile):
                filelist.append(ifile)
            else:
                print(f"ERROR: Can not find {ifile}!")
        return filelist

    def CollectPP(self):
        for k,v in collect_pp(self.pp_path).items():
            if k not in self.pp_dict:
                self.pp_dict[k] = v

    def CollectOrb(self):
        for k,v in collect_pp(self.orb_path).items():
            if k not in self.orb_dict:
                self.orb_dict[k] = v
             
    def CollectPaw(self):
        for k,v in collect_pp(self.paw_path).items():
            if k not in self.paw_dict:
                self.paw_dict[k] = v
            
    def parse_combine_input(self,input_dict, combine_str="|"):
        #parse the combine input, such as "ecutwfc|kspacing": "50|0.2"
        new_input_dict = {}
        for key,value in input_dict.items():
            if combine_str in key and isinstance(value,str):
                keys = key.split(combine_str)
                vs = value.split(combine_str)
                if len(keys) != len(vs):
                    print(f"ERROR: number of keys and values are not equal in {key}={value}")
                    new_input_dict[key] = value
                else:
                    for i in range(len(keys)):
                        new_input_dict[keys[i]] = vs[i]
            else:
                new_input_dict[key] = value
        return new_input_dict
            
    def Construct_input_list(self):
        all_inputs = []
        inputf = None
        if self.example_template != None:
            if os.path.isfile(os.path.join(self.example_template,"INPUT")):
                inputf = os.path.join(self.example_template,"INPUT")
        if self.input_template != None:
            if os.path.isfile(self.input_template):
                inputf = self.input_template 
            else:
                print("WARNING: File %s is not found" % self.input_template)

        print("INPUT: %s" % str(inputf))
        input_constant = {}
        if inputf != None:
            input_constant = PrepareAbacus.ReadInput(inputf)
        
        list_param = {}
        input_constant_common = {}
        for k,v in self.mix_input.items():
            #if value is list type, then we need prepare INPUT for each value
            #input_constant stores the parameters whose value is constant
            #list_param stores the parameters that has several values.
            if isinstance(v,(int,float,str)):
                input_constant[k] = v
                input_constant_common[k] = v
            elif isinstance(v,list):
                if len(v) == 1:
                    input_constant[k] = v[0]
                    input_constant_common[k] = v[0]
                else:    
                    list_param[k] = v
                    if k in input_constant:
                        del input_constant[k]
            else:
                print("WARNING: type of '%s' is" % str(v),type(v),"will not add to INPUT")
                input_constant[k] = v
                
        input_constant = self.parse_combine_input(input_constant) # parse the combine input
        print("Invariant INPUT setting:",str(input_constant))

        all_inputs.append(input_constant)
        for k,v in list_param.items():
            for i,iv in enumerate(v):
                if i == 0:
                    #if v only one value, add in all_inputs directly
                    for iinput in all_inputs:
                        iinput[k] = iv   
                else:
                    #if v has more than two values, deepcopy all_inputs and modify v, and then add to all_inputs
                    if i == 1:  
                        tmp_inputs = copy.deepcopy(all_inputs)
                    for iinput in tmp_inputs:
                        iinput[k] = iv
                    all_inputs += copy.deepcopy(tmp_inputs)
        list_param_new = []
        for k in list_param:
            if "|" in k:
                list_param_new += k.split("|")
            else:
                list_param_new.append(k)
        return [self.parse_combine_input(i) for i in all_inputs],list_param_new

    def Construct_kpt_list(self):
        """
        all_kpt is a list of list (6 element) or filename of KPT.
        If not specify mix_kpt and has specify the KPT template, will retrun a list
        of KPT filename, or will return a null list if no KPT template.
        """
        all_kpt = []

        kptf = None
        if self.example_template != None:
            if os.path.isfile(os.path.join(self.example_template,"KPT")):
                kptf = os.path.join(self.example_template,"KPT")
        if self.kpt_template != None:
            if os.path.isfile(self.kpt_template):
                kptf = self.kpt_template 
            else:
                print("WARNING: File %s is not found" % self.kpt_template)

        if len(self.mix_kpt) == 0:
            if kptf != None:
                all_kpt.append(kptf)
            template_file = kptf
        else:
            for ikpt in self.mix_kpt:
                if isinstance(ikpt,int):
                    all_kpt.append([ikpt,ikpt,ikpt,0,0,0])
                elif isinstance(ikpt,list):
                    if len(ikpt) == 3:
                        all_kpt.append(ikpt+[0,0,0])
                    elif len(ikpt) == 6:
                        all_kpt.append(ikpt)
                    else:
                        print("mix_kpt should be a int or list of 3/6 elements, but not ",ikpt)
                elif isinstance(ikpt,str):
                    allkpt = glob.glob(ikpt)
                    allkpt.sort()
                    for iikpt in allkpt:
                        all_kpt.append(iikpt)
                else:
                    print("mix_kpt should be int or list of 3/6 elements, but not ",ikpt)
            template_file = self.mix_kpt
        print(f"KPT: {template_file}")
        return all_kpt
    
    def Construct_stru_list(self):
        all_stru = []
        struf = None
        if self.example_template != None:
            if os.path.isfile(os.path.join(self.example_template,"STRU")):
                struf = os.path.join(self.example_template,"STRU")
        if self.stru_template != None:
            if os.path.isfile(self.stru_template):
                struf = self.stru_template 
            else:
                print("WARNING: File %s is not found" % self.stru_template)

        if len(self.mix_stru) == 0:
            if struf != None:
                all_stru.append(struf)
            else:
                print("Please set the stru_template!")
            template_file = struf
        else:
            for stru in self.mix_stru:
                allstrus = glob.glob(stru)
                allstrus.sort()
                for istru in allstrus:
                    all_stru.append(os.path.abspath(istru))
                if len(allstrus) == 0:
                    print("Structure file '%s' is not exist" % stru)
            template_file = self.mix_stru
        
        print(f"STRU: {template_file}")
        return all_stru
    
    def prepare(self):
        "Return a dict, and the key is the path of inputs,"
        "and the value is list of structure, kpt, and input setting information."
        "The input setting information is a dict of param name and value"

        if not self.stru_list:
            print("No stru files, skip!!!")
            return None
        
        if not self.kpt_list:
            print("WARNING: not set KPT")
            kpt_list = [None]
        else:
            kpt_list = self.kpt_list
        
        if not self.input_list:
            input_list = [None]
        else:
            input_list = self.input_list 
                   
        ipath = -1
        param_setting = {}
        cwd = os.getcwd()
        stru_num = len(self.stru_list) * len(kpt_list) * len(input_list)
        for istru in self.stru_list:  #iteration of STRU
            stru_data = MyAbacus.AbacusStru.ReadStru(istru)
            if not stru_data:
                print(f"ERROR: read STRU failed in {istru}")
                continue
            stru_path = os.path.split(istru)[0]
            if stru_path == "": stru_path = os.getcwd()
            labels = []
            for ilabel in stru_data.get_label():
                if ilabel not in labels:
                    labels.append(ilabel)
            linkstru = True
            skipstru = False
            allfiles = self.example_template_extra_files + self.extra_files  #files that will be linked
            pp_list = []   #pp file name
            orb_list = []  #orb file name
            paw_list = [] #paw file name
            dpks = None    #dpks file name
            input_in_stru_path = {}
            if os.path.isfile(os.path.join(stru_path,"INPUT")):
                input_in_stru_path = PrepareAbacus.ReadInput(os.path.join(stru_path,"INPUT"))
            print(input_in_stru_path)
            for i,ilabel in enumerate(labels):
                #check pp file 
                if ilabel not in self.pp_dict:
                    #print("label '%s' is found in '%s', but not defined in pp_dict." % (ilabel,istru))
                    if stru_data.get_pp():
                        os.chdir(stru_path)
                        pp_in_stru = os.path.join(input_in_stru_path.get("pseudo_dir",""),stru_data.get_pp()[i])
                        
                        if os.path.isfile(pp_in_stru):
                            print("label '%s': link the pseudopotential file '%s' defined in %s" % (ilabel,pp_in_stru,istru))
                            pp_list.append(os.path.basename(pp_in_stru))
                            allfiles.append(os.path.abspath(pp_in_stru))
                            if pp_list[-1] != stru_data.get_pp()[i]:
                                linkstru = False
                        else:
                            pp_list.append(stru_data.get_pp()[i])
                            print("label '%s': the pseudopotential file '%s' defined in %s is not found" % (ilabel,pp_in_stru,istru))
                            #skipstru = True
                            #os.chdir(cwd)
                            #break
                        os.chdir(cwd)
                else:
                    pp_list.append(os.path.basename(self.pp_dict[ilabel]))  #only store the file name to pp_list
                    allfiles.append(self.pp_dict[ilabel]) #store the whole pp file to allfiles                    
                    if not stru_data.get_pp() or pp_list[-1] != stru_data.get_pp()[i]:
                        linkstru = False

                #check orbital file
                if ilabel not in self.orb_dict:
                    if stru_data.get_orb():
                        #print("label '%s' is found in '%s', but not defined in orb_dict." % (ilabel,istru))
                        os.chdir(stru_path)
                        orb_in_stru = os.path.join(input_in_stru_path.get("orbital_dir",""),stru_data.get_orb()[i])
                        if os.path.isfile(orb_in_stru):
                            print("label '%s': link the orb file '%s' defined in %s" % (ilabel,orb_in_stru,istru))
                            orb_list.append(os.path.basename(orb_in_stru))
                            allfiles.append(os.path.abspath(orb_in_stru))
                            if orb_list[-1] != stru_data.get_orb()[i]:  
                                linkstru = False
                        else:
                            orb_list.append(stru_data.get_orb()[i])
                            print("label '%s': the orbital file '%s' defined in %s is not found." % (ilabel,orb_in_stru,istru))
                        os.chdir(cwd)
                else:
                    orb_list.append(os.path.basename(self.orb_dict[ilabel])) 
                    allfiles.append(self.orb_dict[ilabel])
                    if not stru_data.get_orb() or orb_list[-1] != stru_data.get_orb()[i]:  
                        linkstru = False
                
                #check paw file
                if ilabel not in self.paw_dict:
                    if stru_data.get_paw():
                        os.chdir(stru_path)
                        paw_in_stru = stru_data.get_paw()[i]
                        if os.path.isfile(paw_in_stru):
                            print("label '%s': link the paw file '%s' defined in %s" % (ilabel,paw_in_stru,istru))
                            paw_list.append(os.path.basename(paw_in_stru))
                            allfiles.append(os.path.abspath(paw_in_stru))
                            if paw_list[-1] != stru_data.get_paw()[i]:  
                                linkstru = False
                        else:
                            print("label '%s': the paw file '%s' defined in %s is not found." % (ilabel,paw_in_stru,istru))
                        os.chdir(cwd)
                else:
                    paw_list.append(os.path.basename(self.paw_dict[ilabel])) 
                    allfiles.append(self.paw_dict[ilabel])
                    if not stru_data.get_paw() or paw_list[-1] != stru_data.get_paw()[i]:  
                        linkstru = False

            if skipstru:
                continue
            
            #check dpks 
            if self.dpks_descriptor:
                if os.path.isfile(self.dpks_descriptor):
                    dpks = os.path.basename(self.dpks_descriptor)
                    allfiles.append(self.dpks_descriptor)
                else:
                    print("Error: Can not find file %s, skip the prepare of dpks_descriptor" % self.dpks_descriptor)
                    dpks = None
            else:
                os.chdir(stru_path)
                dpks = stru_data.get_dpks()
                if dpks:
                    if os.path.isfile(dpks):
                        allfiles.append(os.path.abspath(dpks))
                        dpks = os.path.basename(dpks)
                    else:
                        print("Error: deepks descriptor is defined in %s/STRU, but can not find the file, skip the prepare" % stru_path)
                        dpks = None
                else:
                    dpks = None
                os.chdir(cwd)
            if dpks != stru_data.get_dpks():
                linkstru = False

            #stru_data will be writen in new folder
            stru_data.set_pp(pp_list)
            stru_data.set_orb(orb_list)
            stru_data.set_paw(paw_list)
            stru_data.set_dpks(dpks)
            
            for ikpt in kpt_list:  #iteration of KPT 
                for iinput in input_list: #iteration of INPUT
                    if self.pert_stru and self.pert_stru.get("pert_number",0) > 0:
                        stru_data_pert = stru_data.perturb_stru(self.pert_stru.get("pert_number",0),
                                                                cell_pert_frac=self.pert_stru.get("cell_pert_frac",None),
                                                                atom_pert_dist=self.pert_stru.get("atom_pert_dist",None),
                                                                mag_rotate_angle=self.pert_stru.get("mag_rotate_angle",None),
                                                                mag_tilt_angle=self.pert_stru.get("mag_tilt_angle",None),
                                                                mag_norm_dist=self.pert_stru.get("mag_norm_dist",None))
                        linkstru = False
                        create_subpath = True
                    else:
                        stru_data_pert = [stru_data]
                        create_subpath = bool(stru_num > 1)
                    
                    for stru_datai in stru_data_pert:
                        ipath += 1
                        self.write_one_example(create_subpath,ipath,cwd,param_setting,istru,ikpt,iinput,linkstru,stru_datai,allfiles)        
        
        return param_setting

    def write_one_example(self,create_subpath,ipath,cwd,param_setting,stru_path,ikpt,iinput,linkstru,stru_data,allfiles):
        """
        create_subpath: bool, if True, then create subpath for each example. This is True when there are more than one example or more than one inputs/kpt/stru setting.
        ipath: int, the index for example_template
        cwd: str, the current working directory, used to check if the save_path is current path and get the relative path of stru file
        param_setting: dict, to store the param setting of each new abacus job
        stru_path: str, the path of STRU file
        ikpt: KPT setting
        iinput: INPUT setting
        linkstru: bool, if True, then link the STRU file, else copy the STRU file
        stru_data: AbacusStru, the structure data
        allfiles: List, the list of all files that will be linked to the save
        """
        if create_subpath:
                save_path = os.path.join(self.save_path,str(ipath).zfill(5))
        else:
            save_path = self.save_path
        
        if os.path.isdir(save_path) and \
            (not Path(save_path).samefile(cwd)) and \
            (self.bak_file) and \
            (not self.template_is_save_path):
                bk = comm.GetBakFile(save_path)
                shutil.move(save_path,bk)
                
        if not os.path.isdir(save_path):
            os.makedirs(save_path)   
        #store the param setting and path
        param_setting[save_path] = [os.path.relpath(stru_path, cwd),ikpt,{}]
        for input_param in self.input_mix_param:
            param_setting[save_path][-1][input_param] = iinput.get(input_param)
        
        #create INPUT   
        if iinput != None: 
            # if pseudo_dir and orbital_dir is defined in INPUT, then remove them
            iinput.pop("pseudo_dir","")
            iinput.pop("orbital_dir","")  
            PrepareAbacus.WriteInput(iinput,os.path.join(save_path,"INPUT"))
        
        #create KPT
        if ikpt != None:
            kptf = os.path.join(save_path,"KPT")
            if isinstance(ikpt,str):
                if os.path.isfile(kptf) and (not Path(ikpt).samefile(Path(kptf))):
                    os.unlink(kptf)
                
                if not os.path.isfile(kptf):
                    if self.no_link:
                        shutil.copy(os.path.abspath(ikpt),kptf)
                    else:
                        os.symlink(os.path.abspath(ikpt),kptf)
            elif isinstance(ikpt,list):
                PrepareAbacus.WriteKpt(ikpt,kptf)
        
        #create STRU
        struf = os.path.join(os.path.join(save_path,"STRU"))
        if linkstru:
            if os.path.isfile(struf) and (not Path(stru_path).samefile(Path(struf))):
                os.unlink(struf)
                    
            if not os.path.isfile(struf):
                if self.no_link:
                    shutil.copy(os.path.abspath(stru_path),struf)
                else:
                    os.symlink(os.path.abspath(stru_path),struf)
        else:
            stru_data.write(struf)
        
        #link other files
        for ifile in allfiles:
            ifile = os.path.abspath(ifile)
            filename = os.path.basename(ifile)
            target_file = os.path.join(save_path,filename)
            if os.path.isfile(target_file) and not Path(ifile).samefile(Path(target_file)):
                os.unlink(target_file)
            if not os.path.isfile(target_file):
                if self.no_link:
                    shutil.copy(ifile,target_file)
                else:
                    os.symlink(ifile,target_file)  
    
    @staticmethod
    def WriteKpt(kpoint_list:List = [1,1,1,0,0,0],file_name:str = "KPT", model:str = "gamma"):
        MyAbacus.WriteKpt(kpoint_list,file_name,model)
    
    @staticmethod
    def ReadInput(INPUTf: str = None, input_lines: str = None) -> Dict[str,any]:
        return MyAbacus.ReadInput(INPUTf,input_lines)

    @staticmethod
    def WriteInput(input_context:Dict[str,any],
                   INPUTf: str = "INPUT"):
        MyAbacus.WriteInput(input_context,INPUTf)           

def CheckExample(example_path,description:str=""):
    print(f"Check ABACUS inputs: {example_path} {description}")
    if not os.path.isdir(example_path):
        print("Can not find example path %s" % example_path)
        return False
    if not os.path.isfile(os.path.join(example_path,"INPUT")):
        print("Can not find INPUT file in %s" % example_path)
        return False
    if not os.path.isfile(os.path.join(example_path,"STRU")):
        print("Can not find STRU file in %s" % example_path)
        return False
    iinput = PrepareAbacus.ReadInput(os.path.join(example_path,"INPUT"))
    pp_path = iinput.get("pseudo_dir","")
    orb_path = iinput.get("orbital_dir","")
    paw_path = iinput.get("paw_dir","")
    use_paw = iinput.get("use_paw",False)
    if use_paw and use_paw.isdigit() and int(use_paw):
        use_paw = True
    else:
        use_paw = False
    istru = MyAbacus.AbacusStru.ReadStru(os.path.join(example_path,"STRU"))
    if not istru:
        print("Read STRU file failed in %s" % example_path)
        return False
    
    cwd = os.getcwd()
    allpass = True
    os.chdir(example_path)
    #check pp
    if not use_paw:
        ppfiles = istru.get_pp()
        if not ppfiles:
            print("Can not find PP in %s" % example_path)
            allpass = False
        else:
            for ppfile in ppfiles:
                real_ppfile = os.path.join(pp_path,ppfile)
                if not os.path.isfile(real_ppfile):
                    print("Can not find PP file %s in %s" % (real_ppfile,example_path))
                    allpass = False
    else:
        pawfiles = istru.get_paw()
        if not pawfiles:
            print("Can not find PAW in %s" % example_path)
            allpass = False
        else:
            for pawfile in pawfiles:
                real_pawfile = os.path.join(paw_path,pawfile)
                if not os.path.isfile(real_pawfile):
                    print("Can not find PAW file %s in %s" % (real_pawfile,example_path))
                    allpass = False
    
    #check orb
    basis = iinput.get("basis_type","pw").lower()
    if basis != "pw" and basis.startswith("lcao"):
        orbfiles = istru.get_orb()
        if not orbfiles:
            print("Can not find ORB in %s/STRU" % example_path)
            allpass = False
        else:
            for orbfile in orbfiles:
                real_orbfile = os.path.join(orb_path,orbfile)
                if not os.path.isfile(real_orbfile):
                    print("Can not find ORB file %s in %s" % (real_orbfile,example_path))
                    allpass = False
    
    #check dpks
    deepks_scf = iinput.get("deepks_scf",False)
    if deepks_scf:
        dpks_model = iinput.get("deepks_model",None)
        if not dpks_model:
            print("deepks_model is not defined in %s/INPUT" % example_path)
            allpass = False
        else:
            if not os.path.isfile(dpks_model):
                print("Can not find dpks model file %s in %s" % (dpks_model,example_path))
                allpass = False
    deepks_out_labels = iinput.get("deepks_out_labels",None)
    if deepks_scf or deepks_out_labels:
        dpks_descriptor = istru.get_dpks()
        if not dpks_descriptor:
            print("Read dpks descriptor failed in %s/STRU" % example_path)
            allpass = False
        elif not os.path.isfile(dpks_descriptor):
            print("Can not find dpks descriptor file %s in %s" % (dpks_descriptor,example_path))
            allpass = False
    
    # check KPT
    # only basis is lcao and use gamma_only, or has define kspacing, KPT file is not needed
    gamma_only = iinput.get("gamma_only",False)
    kspacing = iinput.get("kspacing",None)
    if not ((basis.startswith("lcao") and gamma_only) or kspacing): 
        if not os.path.isfile("KPT"):
            print("Can not find KPT file in %s" % example_path)
            allpass = False 
    os.chdir(cwd)
    return allpass 

def CommPath(pathlist):
    "return (commpath pathlist_without_commpath)"
    if len(pathlist) == 0:
        return None,pathlist
    commpath = os.path.commonpath(pathlist)
    if commpath != "":
        length = len(commpath) + 1
    else:
        length = 0
    return commpath,[i[length:] for i in pathlist]

def DoPrepare(param_setting: Dict[str, any], save_folder: str, no_link: bool = False) -> List[Dict[str, dict]]:  
    """
    param_setting is a dictionary like:
    {
        "example_template":["example_path"]
        "input_template":"INPUT",
        "kpt_template":"KPT",
        "stru_template":"STRU",
        "strus": str/list,  # the stru file of other format, like cif, vasp, etc.
        "stru_format": "cif",  # the format of stru file, should be cif or dpdata supportted format, if strus is setted, then example_template is not needed.
        "mix_input":{
            "ecutwfc":[50,60,70],
            "kspacing":[0.1,0.12,0.13]
        },
        "mix_kpt":[],
        "mix_stru":[],
        "pert_stru":{
            "pert_number": 0,
            "cell_pert_frac": null,
            "atom_pert_dist": null,
            "mag_rotate_angle": null,
            "mag_tilt_angle": null,
            "mag_norm_dist": null
            },
        "pp_dict":{},
        "orb_dict":{},
        "pp_path": str,
        "orb_path": str,
        "dpks_descriptor":"",
        "extra_files":[],
        "link_example_template_extra_files":true,
        "bak_file":true,
        "abacus2qe": bool,
        "qe_setting":{},
        "abacus2vasp": bool,
        "potcar": str,
        "vasp_setting":{},
        "abacus2cp2k": bool,
        "cp2k_setting":{},
        "mix_input_comment":"Do mixing for several input parameters. The diffrent values of one parameter should put in a list.",
        "mix_kpt_comment":"If need the mixing of several kpt setting. The element should be an int (eg 2 means [2,2,2,0,0,0]), or a list of 3 or 6 elements (eg [2,2,2] or [2,2,2,0,0,0]).",
        "mix_stru_comment":"If need the mixing of several stru files. Like: ["a/stru","b/stru"], 
        "bak_file_comment":"If need to bak the old folder if the new folder is exist.",      
    },

    save_folder specifies the destination of the new created examples.
    
    if strus and stru_format is setted, then will firstly convert it to ABACUS format, and save to new folder named by %06d.

    Return a list of dict, which is related to each example_template element. The key of the dict is the newcreated example path, and the value
    is the STRU/KPT/INPUT settings. 
    """
    strus = param_setting.get("strus",None)
    stru_format = param_setting.get("stru_format",None)
    if strus is not None:
        assert stru_format is not None, "If strus is setted, then stru_format should be setted."
        example_template = translate_strus(strus,stru_format)
        if example_template is None:
            print("Convert strus to ABACUS format failed.")
            return []
        if param_setting.get("example_template",None) is not None:
            print("WARNING: strus is setted, example_template is not used.")
    else:
        example_template = param_setting.get("example_template",None)
        
    if isinstance(example_template,str):
        example_template = glob.glob(example_template)
        # if __MACOSX folder is exist, then remove it
        if "__MACOSX" in example_template:
            example_template.remove("__MACOSX")
        if "__MACOSX/" in example_template:
            example_template.remove("__MACOSX/")
        example_template.sort()
    elif isinstance(example_template,list):
        tmp = copy.deepcopy(example_template)
        example_template = []
        for i in tmp:
            alli = glob.glob(i)
            # if __MACOSX folder is exist, then remove it
            if "__MACOSX" in alli:
                alli.remove("__MACOSX")
            if "__MACOSX/" in alli:
                alli.remove("__MACOSX/")
            alli.sort()
            example_template += alli
    else:
        example_template = [example_template]
    
    all_path_setting = []
    commpath,example_template_nocomm = CommPath(example_template)
    print(commpath,example_template_nocomm)
    for idx,iexample in enumerate(example_template):
        if not os.path.isdir(iexample):
            print("Can not find folder %s" % iexample)
            continue
        #if len(example_template) > 1:
        #    save_path = os.path.join(save_folder,example_template_nocomm[idx])
        #    print("\n%s" % iexample)
        #else:
        #    if Path(save_folder) == Path("."):
        #        save_path = commpath
        #    else:
        #        save_path = save_folder
        save_path = os.path.join(save_folder,iexample)
        prepareabacus = PrepareAbacus(save_path=save_path,
                                  example_template=iexample,
                                  input_template=param_setting.get("input_template",None),
                                  kpt_template=param_setting.get("kpt_template",None),
                                  stru_template=param_setting.get("stru_template",None),
                                  mix_input=param_setting.get("mix_input",{}),
                                  mix_kpt=param_setting.get("mix_kpt",[]),
                                  mix_stru=param_setting.get("mix_stru",[]),
                                  pert_stru=param_setting.get("pert_stru",{}),
                                  pp_dict=param_setting.get("pp_dict",{}),
                                  orb_dict=param_setting.get("orb_dict",{}),
                                  pp_path=param_setting.get("pp_path",None),
                                  orb_path=param_setting.get("orb_path",None),
                                  dpks_descriptor=param_setting.get("dpks_descriptor",None),
                                  extra_files=param_setting.get("extra_files",[]),
                                  bak_file = param_setting.get("bak_file",True),
                                  no_link=no_link,
                                  link_example_template_extra_files=param_setting.get("link_example_template_extra_files",True)
                                  )
        all_path_setting.append(prepareabacus.prepare())
            
    if param_setting.get("abacus2qe",False):
        print("\nConvert ABACUS inputs to QE inputs")
        for isetting in all_path_setting:
            if isetting:
                for ipath in list(isetting.keys()):
                    print(ipath)
                    try:
                        Aba2Qe.Abacus2Qe(ipath,save_path=os.path.join(ipath,"input"),qe_param=param_setting.get("qe_setting",{}))
                    except:
                        traceback.print_exc()
    
    if param_setting.get("abacus2vasp",False):
        print("\nConvert ABACUS inputs to VASP inputs")
        potcar = param_setting.get("potcar",None)
        vasp_setting = param_setting.get("vasp_setting",{})
        for isetting in all_path_setting:
            if isetting:
                for ipath in list(isetting.keys()):
                    print(ipath)
                    try:
                        Aba2Vasp.Abacus2Vasp(ipath,save_path=ipath,potcar=potcar,vasp_setting=vasp_setting)
                    except:
                        traceback.print_exc()
    
    if param_setting.get("abacus2cp2k",False):
        print("\nConvert ABACUS inputs to CP2K inputs")
        cp2k_setting = param_setting.get("cp2k_setting",{})
        for isetting in all_path_setting:
            if isetting:
                for ipath in list(isetting.keys()):
                    print(ipath)
                    try:
                        Aba2Cp2k.Abacus2Cp2k(ipath,save_path=os.path.join(ipath,"cp2k.inp"),cp2k_setting=cp2k_setting)
                    except:
                        traceback.print_exc()
                      
    return all_path_setting

def PrepareInput(param):
    param_setting_file = param.param
    save_folder = param.save
    if not os.path.isfile(param_setting_file):
        print("Can not find file %s!!!" % param_setting_file)
        sys.exit(1)

    param_setting = json.load(open(param_setting_file))
    if "prepare" in param_setting:
        param_setting = param_setting["prepare"]
    
    all_path_setting = DoPrepare(param_setting,save_folder,param.nolink)
    for path_setting in all_path_setting:
        if path_setting:
            for k,v in path_setting.items():
                print("%s:%s" % (k,str(v)))

def PrepareArgs(parser):  
    parser.description = "This script is used to prepare the INPUTS OF ABACUS JOB"
    parser.add_argument('-p', '--param', type=str, help='the parameter file, should be .json type',required=True)
    parser.add_argument('-s', '--save', type=str,  default="abacustest",help='where to store the inputs, default is abacustest ')
    parser.add_argument('--nolink',  nargs='?',type=int, const=1, default=0,help='if link the files in the example folder, default is 0')
    return parser

def main():
    parser = argparse.ArgumentParser()
    param = PrepareArgs(parser).parse_args()
    PrepareInput(param)
    
if __name__ == "__main__":
    main()



from . import globV
import os,shutil,glob,json
from pathlib import Path

def printinfo(istr,*args):
    LOGFILE = "abacustest.log"
    output = " ".join([str(istr)]+[str(i) for i in args])
    with open(LOGFILE,'a+') as f1:
        f1.write(output + "\n")
    print(output,flush=True)
        
def GetBakFile(sfile):
    while sfile[-1] == '/':
        sfile = sfile[:-1]
    n = 1
    bk = sfile + ".bak%d" % n
    while os.path.exists(bk):
        n += 1
        bk = sfile + ".bak%d" % n
    return bk

def CopyFiles(path1,path2,move = False):
    '''copy the files in path1 to path2'''
    abspath1 = os.path.abspath(path1)
    abspath2 = os.path.abspath(path2)
    
    if not os.path.isdir(abspath2):
        os.makedirs(abspath2)
    if not os.path.isdir(abspath1):
        print(f"{abspath1} is not a exist path")
        return
    
    def CopyFile(old_path,new_path):
        for ifile in os.listdir(old_path):
            iold_path = os.path.join(old_path,ifile)
            if os.path.isfile(iold_path):
                shutil.copy(iold_path,new_path)
            else:
                if not os.path.isdir(os.path.join(new_path,ifile)):
                    os.makedirs(os.path.join(new_path,ifile))
                CopyFile(iold_path,os.path.join(new_path,ifile))

    if abspath2.startswith(abspath1):
        '''
        If path2 is a son path of path1,
        we will firstly create a tmp path, and copy/move
        files in path1 to the tmp path, and then replace tmp path
        to path2.
        '''
        tmp_path = GetBakFile(os.path.split(abspath1)[0])
        os.makedirs(tmp_path)
        if move:
            for i in os.listdir(abspath1):
                shutil.move(os.path.join(abspath1,i),tmp_path)
        else:
            #shutil.copytree(abspath1,tmp_path,dirs_exist_ok=True)
            CopyFile(abspath1,tmp_path)
               
        #shutil.copytree(tmp_path,abspath2,dirs_exist_ok=True)
        CopyFile(tmp_path,abspath2)
        shutil.rmtree(tmp_path)
        
    else:
        if os.path.isfile(abspath1):
            if move:
                shutil.move(abspath1,abspath2)
            else:
                shutil.copy(abspath1,abspath2)
            return
        elif os.path.isdir(abspath1):
            if move:
                for i in os.listdir(abspath1):
                    shutil.move(os.path.join(abspath1,i),abspath2)
            else:
                #shutil.copytree(abspath1,abspath2,dirs_exist_ok=True)
                CopyFile(abspath1,abspath2)

def LinkFiles(src_list, dst_path):
    # the src_list should be a list of relative path, and the dst_path should be a absolute path
    # will create the soft link in dst_path with the same path tree as src_list
    for src in src_list:
        src = src.rstrip("/")
        src_root, src_name = os.path.split(src)   
        if os.path.isfile(os.path.join(dst_path,src)):
            os.remove(os.path.join(dst_path,src))
        elif os.path.isdir(os.path.join(dst_path,src)):
            shutil.rmtree(os.path.join(dst_path,src))
                        
        if not os.path.exists(os.path.join(dst_path,src_root)):
            os.makedirs(os.path.join(dst_path,src_root))
        os.symlink(os.path.abspath(src),os.path.join(dst_path,src))           

def CollectFileName(paths):
    #Recursively find all files under the paths
    allfiles = []
    for ipath in glob.glob(os.path.join(paths,"*")):
        if os.path.isfile(ipath):
            allfiles.append(ipath)
        elif os.path.isdir(ipath):
            allfiles += CollectFileName(ipath)
    return allfiles            

def ProduceExecutor(param,group_name="abacustesting"):
    from dflow.plugins.bohrium import BohriumContext, BohriumExecutor
    from dflow.plugins.dispatcher import DispatcherExecutor
    from dflow.plugins.bohrium import TiefblueClient,create_job_group
    
    if "bohrium" in param and param["bohrium"]:
        bohrium_set = {}
        for key in param["bohrium"]:
            if key == 'scassType':
                bohrium_set['scass_type'] = param["bohrium"][key]
            elif key == 'jobType':
                bohrium_set['job_type'] = param["bohrium"][key]
            else:
                bohrium_set[key] = param["bohrium"][key]

        if 'platform' not in bohrium_set:
            bohrium_set['platform'] = 'ali'

        bohrium_set["bohr_job_group_id"] = create_job_group(group_name)
        
        if not globV.get_value("BOHRIUM_EXECUTOR"):    
            dispatcher_executor = DispatcherExecutor(
                machine_dict={
                    "batch_type": "Bohrium",
                    "context_type": "Bohrium",
                    "remote_profile": {"input_data": bohrium_set},
                    },
                image_pull_policy = "Always",
                retry_on_submission_error=3,
                #image=param["bohrium"].get("dispatcher_image","registry.dp.tech/public/dptechnology/dpdispatcher:v0.6.0"),
            )
            if "dispatcher_image" in param["bohrium"]:
                dispatcher_executor.image = param["bohrium"]["dispatcher_image"]
            #comm.printinfo("set bohrium: %s"%str(bohrium_set))
            return dispatcher_executor,bohrium_set
        else:
            executor = BohriumExecutor(
                executor= "bohrium_v2",
                extra={
                    "scassType": bohrium_set['scass_type'],
                    "platform": bohrium_set['platform'] ,
                    "projectId": globV.get_value("PRIVATE_SET").get("bohrium_project_id"),
                    "jobType":  bohrium_set['job_type']
                }
            )
            return executor,bohrium_set
    elif "dispatcher" in param and param["dispatcher"]:
        '''
        host: remote host
        queue_name: queue name
        port: SSH port
        username: username
        private_key_file: private key file for SSH
        image: image for dispatcher
        command: command for dispatcher
        remote_command: command for running the script remotely
        map_tmp_dir: map /tmp to ./tmp
        machine_dict: machine config for dispatcher
        resources_dict: resources config for dispatcher
        task_dict: task config for dispatcher
        json_file: JSON file containing machine and resources config


        dispatcher_executor = DispatcherExecutor(
            host = param["dispatcher"].get("host",None),
            queue_name = param["dispatcher"].get("queue_name",None),
            port = param["dispatcher"].get("port",22),
            username = param["dispatcher"].get("username","root"),
            private_key_file = param["dispatcher"].get("private_key_file",None),
        #    image = image,
        #    command = command,
        #    remote_command = remote_command,
        #    map_tmp_dir = map_tmp_dir,
            machine_dict = param["dispatcher"].get("machine_dict",None),
            resources_dict = param["dispatcher"].get("resources_dict",None),
            task_dict = param["dispatcher"].get("task_dict",None),
            json_file = param["dispatcher"].get("json_file",None)
            )
        '''
        dispatcher_executor = DispatcherExecutor(**param["dispatcher"])
        import copy
        tmp_param = copy.deepcopy(param["dispatcher"])
        hide_config_in_dispatcher(tmp_param)
        return dispatcher_executor,tmp_param
    else:
        return None,None

def hide_config_in_dispatcher(tmp_param):
    # tmp_param is the value of dispatcher
    if "host" in tmp_param:
        tmp_param["host"] = "******"
    if "username" in tmp_param:
        tmp_param["username"] = "******"
    if "port" in tmp_param:
        tmp_param["port"] = "******"
    if "private_key_file" in tmp_param:
        tmp_param["private_key_file"] = "******"
    if "machine_dict" in tmp_param and "remote_profile" in tmp_param["machine_dict"]:
        if "hostname" in tmp_param["machine_dict"]["remote_profile"]:
            tmp_param["machine_dict"]["remote_profile"]["hostname"] = "******"
        if "username" in tmp_param["machine_dict"]["remote_profile"]:
            tmp_param["machine_dict"]["remote_profile"]["username"] = "******"
        if "password" in tmp_param["machine_dict"]["remote_profile"]:
            tmp_param["machine_dict"]["remote_profile"]["password"] = "******"
        if "port" in tmp_param["machine_dict"]["remote_profile"]:
            tmp_param["machine_dict"]["remote_profile"]["port"] = "******"

def FindLocalExamples_new(example,only_folder=False,oneartifact=False):
    from dflow import upload_artifact
    #use glob.glob find all examples, and transfer to artifact
    #example = [[*],*]    
    examples_name = []  
    examples_name1 = []
    for i in example:
        if isinstance(i,list):
            example_tmp = []
            for j in i:
                example_tmp += glob.glob(j)
            if len(example_tmp) > 0:
                example_tmp.sort()
                tmp = []
                for ii in example_tmp:
                    # if ii is __MACOSX, skip it
                    if os.path.split(ii)[-1].startswith("__MACOSX"):
                        continue
                    if only_folder and not os.path.isdir(ii):
                        continue
                    tmp.append(ii)
                    examples_name1.append(ii)
                if len(tmp) > 0:
                    tmp.sort()
                    examples_name.append(tmp)
        elif isinstance(i,str):
            for ii in glob.glob(i):
                if os.path.split(ii)[-1].startswith("__MACOSX"):
                    continue
                if only_folder and not os.path.isdir(ii):
                        continue
                examples_name.append([ii])
                examples_name1.append(ii)
        else:
            printinfo(i,"element of 'example' should be a list, or str")
    
    examples = None
    if len(examples_name) > 0:  
        if oneartifact:
            examples_name = list(set(examples_name1))
            examples_name.sort()
            examples = [[upload_artifact(examples_name,archive=globV.get_value("COMPRESS"))]]
            examples_name = [examples_name]
        else:
            examples = []
            examples_name.sort()
            for ii in examples_name:
                examples.append([upload_artifact(iii,archive=globV.get_value("COMPRESS")) for iii in ii])    
    
    # if oneartifact is True, examples = [[artifact]], examples_name = [[example1,example2,...]]
    # else examples = [[artifact1,artifact2,...],[artifact3,artifact4,...],...], examples_name = [[example1,example2,...],[example3,example4,...],...]
    return examples,examples_name

def transfer_source_to_artifact(example,source=None,source_type="local",only_folder=True,oneartifact=False):
    # example specify the example name
    # if onearitfact is True, then the source will be transfered to one artifact
    # else, the source will be transfered to a list of artifacts (for rundft)
    # for rundft, need set only_folder = True
    if isinstance(example,str):
        example = [example]
    
    examples,examples_name = [],[]   
     
    if source_type == "local":
        examples,examples_name = FindLocalExamples_new(example,only_folder=only_folder,oneartifact=oneartifact)
    
    # examples = [[artifact1,artifact2,...],[artifact3,artifact4,...],...]
    # examples_name = [[example1,example2,...],[example3,example4,...],...]
    return examples,examples_name

def SplitGroup(examples,examples_name,ngroup): 
    #examples = [*[*],*]
    newexamples = [] 
    newexamples_name = []
    se = 0
    mod = len(examples) % ngroup
    for i in range(ngroup):
        example_tmp = []
        example_name_tmp = []
        add = 1 if mod > 0 else 0 
        ee = se + int(len(examples)/ngroup) + add
        for ie in range(se,ee):
            example_tmp += examples[ie]
            example_name_tmp += examples_name[ie]
        newexamples.append(example_tmp)
        newexamples_name.append(example_name_tmp)
        
        if mod > 0: mod -= 1
        se = ee
    return newexamples,newexamples_name

def SplitGroupSize(examples,examples_name,group_size):
    #examples = [*[*],*]
    newexamples = [[]] 
    newexamples_name = [[]]
    
    i = 0
    j = 0
    while i < len(examples):
        newexamples[-1] += examples[i]
        newexamples_name[-1] += examples_name[i]
        j += 1
        i += 1
        if i >= len(examples):
            break
        if j >= group_size:
            j = 0
            newexamples.append([])
            newexamples_name.append([])
    return newexamples,newexamples_name

def ParseSavePath(save_path):
    if save_path != None:
        save_path = save_path.strip()
        if save_path == '':
            save_path = globV.get_value("RESULT")
    else:
        save_path = globV.get_value("RESULT")
    return save_path

def ParseSubSavePath(sub_save_path):
    if sub_save_path == None:
        sub_save_path = ""
    elif isinstance(sub_save_path,str):
        sub_save_path = sub_save_path.strip()
        if Path(sub_save_path) == Path("."):
            sub_save_path = ""
    else:
        printinfo("the type of 'sub_save_path' should be 'str', but not '%s'. %s" % (type(sub_save_path),str(sub_save_path)))
        sub_save_path = ""
    return sub_save_path 

def SetEnvs():
    if globV.get_value("PRIVATE_SET",None) == None:
        return None
    from dflow import Secret
    envs = {}
    for k,v in globV.get_value("PRIVATE_SET").items():
        if isinstance(v,dict):
            envs[k.upper()] = Secret("'" + json.dumps(v) + "'")
        else:
            envs[k.upper()] = Secret(str(v))
    return envs

def run_command(
        cmd,
        shell = True
):
    import subprocess,select
    process = subprocess.Popen(
        cmd, 
        stdout=subprocess.PIPE, 
        stderr=subprocess.PIPE,
        shell=shell,
        executable='/bin/bash'
    )
    out = ""
    err = ""
    while True:
        # 监视stdout和stderr文件描述符的可读状态
        readable, _, _ = select.select([process.stdout, process.stderr], [], [])

        # 读取已经准备好的输出
        for fd in readable:
            if fd == process.stdout:
                line = process.stdout.readline()
                print(line.decode()[:-1])
                out += line.decode()
            elif fd == process.stderr:
                line = process.stderr.readline()
                if line.decode()[:-1].strip() != "":
                    print("STDERR:", line.decode()[:-1])
                err += line.decode()

        # 如果子进程已经结束，则退出循环
        return_code = process.poll()
        if return_code is not None:
            break
    return return_code, out, err




from dp.launching.typing.basic import BaseModel, Int, String, Float, List, Optional, Union, Dict
from dp.launching.typing import InputFilePath, OutputDirectory
from dp.launching.typing import (
    BaseModel,
    Set,
    Boolean,
    Field,
    DflowAccessToken,
    DflowArgoAPIServer,
    DflowK8sAPIServer,
    DflowStorageEndpoint,
    DflowStorageRepository,
    BohriumMachineType,
    BohriumImage,
    BohriumPlatform,
    BohriumJobType,
    BohriumUsername,
    BohriumPassword,
    BohriumProjectId,
    BohriumTicket,
    BenchmarkLabels,
    BenchmarkTags,
    DflowLabels 
)
from typing import Literal

import re,os

class OutputSet(BaseModel):
    IO_output_path: OutputDirectory = Field(default="./outputs")

class ConfigSet(BaseModel):
    #Bohrium config
    Config_bohrium_username:   BohriumUsername
#    Config_bohrium_password:   BohriumPassword
    Config_bohrium_project_id:     BohriumProjectId
    Config_bohrium_ticket: BohriumTicket

    #dflow set
    Config_dflow_host: DflowArgoAPIServer
    Config_dflow_s3_config_endpoint: DflowStorageEndpoint
    Config_dflow_k8s_api_server: DflowK8sAPIServer
    Config_dflow_token: DflowAccessToken
    Config_dflow_labels: DflowLabels

class ConfigSetGithub(BaseModel):
    Config_github_username: String = Field(default="",title="github_username",description="github username")
    Config_github_email: String = Field(default="",title="github_email",description="github email")
    Config_github_token: String = Field(default="",title="github_token",description="github token")

class ImageDispatcher(BaseModel):
    type: Literal["Use dp-dispatcher"]

    '''
    "machine_dict": {
                    "remote_root": "xxxx",
                    "remote_profile": {
                        "hostname": "xxxx",
                        "username": "xxx",
                        "password": "xxx",
                        "port": 22
                    }
    },
    "resources_dict": {
        "number_node": 1,
        "cpu_per_node": 8,
        "gpu_per_node": 1,
        "queue_name": "normal"
    }
    '''
    host: String = Field(default="",title="host",description="host of remote server")
    username: String = Field(default="",title="username",description="username of remote server")
    password: String = Field(default="",title="password",description="password of remote server")
    port: Int = Field(default=22,title="port",description="port of remote server")
    remote_root: String = Field(default="",title="remote_root",description="an absolute path on remote server to run the calculation")
    number_node: Int = Field(default=1,title="number_node",description="number of nodes to run the calculation")
    cpu_per_node: Int = Field(default=8,title="cpu_per_node",description="number of cpus per node to run the calculation")
    gpu_per_node: Int = Field(default=0,title="gpu_per_node",description="number of gpus per node to run the calculation")
    queue_name: String = Field(default="",title="queue_name",description="queue name to run the calculation")

    @classmethod
    def construct_dispatcher(cls,opts):
        param = {
            "dispatcher":{
                "machine_dict":{
                    "remote_root":opts.remote_root,
                    "remote_profile":{
                        "hostname":opts.host,
                        "username":opts.username,
                        "password":opts.password,
                        "port":opts.port
                    }
                },
                "resources_dict":{
                    "number_node":opts.number_node,
                    "cpu_per_node":opts.cpu_per_node,
                    "gpu_per_node":opts.gpu_per_node,
                    "queue_name":opts.queue_name
                }
            }
        }
        if opts.gpu_per_node < 1:
            del param["dispatcher"]["resources_dict"]["gpu_per_node"]
        if opts.queue_name.strip() == "":
            del param["dispatcher"]["resources_dict"]["queue_name"]
        return param
    
class NgroupSet(BaseModel):
    ngroup: Int = Field(
        default=0, description="Number of groups to run in parallel. If not set, all examples will be run in parallel.", ge=0)

class TrackingSet(BaseModel):
    Tracking_metrics: Boolean = Field(
        default=False, description="If tracking, will display historical metrics values based on test and experience")
    Tracking_token: String = Field(
        default=None, description="If want to track metrics, please enter your token to access AIM")
    Tracking_tags: BenchmarkTags
    #tags = [f"benchmark-application-{application.name}", f"benchmark-version-{job.version}", f"benchmark-schedule-{job.properties.get('source_name', 'none')}",
    #        f"benchmark-job-{job.name}"]

    @classmethod
    def parse_obj(cls, opts):
        if opts.Tracking_metrics:
            default_tags = opts.Tracking_tags
            schedule = "_".join(
                re.split("-", default_tags[2], 2)[-1].strip().split())
            application_name = re.split("-", default_tags[0], 2)[-1]
            job_name = re.split("-", default_tags[3], 2)[-1]
            if opts.Tracking_token != None and opts.Tracking_token.strip() != "":
                token = opts.Tracking_token
            else:
                token = os.environ.get("AIM_ACCESS_TOKEN")

            return {
                "name": schedule + "." + job_name,
                "experiment": application_name + "/benchmark",
                "tags": default_tags,
                "token": token
            }
        else:
            return None


class myLog:
    def __init__(self):
        self.logs = ""
    
    def __add__(self, other):
        if isinstance(other, myLog):
            self.logs += other.logs
        else:
            self.logs += str(other)
        return self

    def iprint(self, mess, *args):
        allmess = " ".join([str(mess)]+[str(i) for i in args])
        print(allmess)
        self.logs += allmess + "\n"

    def write(self, filename):
        with open(filename, 'w') as f1:
            f1.write(self.logs)



import os,json,glob,shutil,traceback
import subprocess,copy
from abacustest.lib_prepare.abacus import AbacusStru,ReadInput,ReadKpt
import select
from typing import List, Dict, Union, Optional, Tuple, Literal
from abacustest.lib_prepare.comm import IsTrue

BOHRIUM_DES = '''If you use Bohrium to accelerate the calculation, you need to set below environment variables:
    export BOHRIUM_USERNAME=<your username> BOHRIUM_PASSWORD=<your password> BOHRIUM_PROJECT_ID=<your project id> 
Or you can add below block in your setting.json:
    "config": {
        "bohrium_username": "<your username>",
        "bohrium_password": "<your password>",
        "bohrium_project_id": "<your project id>"
    } 
Or you can use dispatcher to submit the calculation to the remote server,
and please refer to the dispatcher section for more details in https://github.com/pxlxingliang/abacus-test/tree/develop.'''

def doc_after_prepare(model_name, jobs, files,has_prepare=True):
    print(f"\nPrepare the {model_name} successfully!!!")
    print("\nYou can choose one of below methods to do the calculation:")
    print("1. run with abacustest locally")
    print(BOHRIUM_DES,)
    print("After finishing the setting, you can run below command to submit the calculation:")
    print("        abacustest submit -p setting.json &")
    print("\n2. run with bohrium app abacustest:")
    print("Compress the inputs and submit to https://app.bohrium.dp.tech/abacustest/?request=GET%3A%2Fapplications%2Fabacustest with Reuse Submodel.")
    print("Compress command:")
    print("    tar -zcvf %s.tar.gz %s" % (model_name, " ".join(jobs+files)))
    if has_prepare:
        print("\n3. run with your own script:")
        print("You can execute below command and abacustest will generate all inputs in folder abacustest/, and then you can submit the calculation by yourself.")
        outfolder = "" if len(jobs) > 1 else f"-s {jobs[0]}"
        print("    abacustest prepare -p setting.json " + outfolder + "\n\n")
    
def dump_setting(setting):
    """
    Dump the given setting dictionary to a JSON file named 'setting.json'.
    If 'setting.json' already exists, it will be backed up with a numbered suffix.

    Args:
        setting (dict): The setting dictionary to be dumped.

    Returns:
        None
    """
    if os.path.isfile("setting.json"):
        i = 1
        settingbakf = f"setting.json.bak{i}"
        while os.path.isfile(settingbakf):
            i += 1
            settingbakf = f"setting.json.bak{i}"
        print("Warning: setting.json exists. Will backup it to", settingbakf)
        os.system(f"mv setting.json {settingbakf}")
    json.dump(setting, open("setting.json", "w"), indent=4)

def get_physical_cores():
    """
    Get the number of physical cores in the system.

    Returns:
        int: The number of physical cores.

    Raises:
        CalledProcessError: If the command execution fails.
    """
    cmd = "lscpu | grep 'Core(s) per socket:' | awk '{print $4}'"
    cores = subprocess.check_output(cmd, shell=True).decode().strip()
    return int(cores)

def get_job_list(jobs):
    '''
    Get a list of job directories from the given list of job patterns.

    Args:
        jobs (list): A list of job patterns.

    Returns:
        list: A list of job directories.
    '''
    job_list = []
    for job in jobs:
        for ijob in glob.glob(job):
            if os.path.isdir(ijob):                        
                job_list.append(ijob)
    return job_list
    
def gen_supermetrics(file_name, sm_name=None, image=True):
    # file_type should be image or html
    if sm_name == None:
        basename = os.path.basename(file_name.rstrip("/"))
        if "." in basename:
            sm_name = os.path.splitext(basename)[0]
        else:
            sm_name = basename
    return {
        sm_name: {"type": {True:"image", False: "html"}.get(bool(image)),
                  "file": file_name}
    } 

def bak_file(filename, bak_org=False):
    '''
    Check if the file exists, and if so:
    if bak_org is True, backup the original file to a non-exist filename.bak{n}, n = 1,2,3,...
    else, return the filename.bak{n} that does not exist.
    '''
    if not os.path.exists(filename):
        return filename
    
    i = 1
    bakf = f"{filename}.bak{i}"
    while os.path.exists(bakf):
        i += 1
        bakf = f"{filename}.bak{i}"
        
    if bak_org:
        os.system(f"mv {filename} {bakf}")
        return filename 
    else:
        return bakf

def clean_files(ipath, f_list=[], folder_list=[]):
    '''
    Clean the files in the given folder.
    '''
    for ifile in f_list:
        for jfile in glob.glob(os.path.join(ipath,ifile)):
            if os.path.isfile(jfile):
                os.remove(jfile)
    for ifolder in folder_list:
        for jfolder in glob.glob(os.path.join(ipath,ifolder)):
            if os.path.isdir(jfolder):
                shutil.rmtree(jfolder)       

def get_abacus_inputfiles(ipath):
    """
    Find the input files in the given path and return a dict:
    {
        "input": input_param, # a dict of the input parameters, kpt_file, stru_file, pseudo_dir, orb_dir will be popped out or modified
        "stru": stru, # an AbacusStru object or None. The pp/orb now is only the base name.
        "kptf": kptf, # the absolute path of the kpt file or None if not found
        "pp": pp_abs_path, # a list of the absolute path of the pseudopotential files or [] if defined in STRU
        "orb": orb_abs_path, # a list of the absolute path of the orbital files or [] if defined in STRU
        "extra_files": extra_files # a list of the absolute path of the extra files
    }
    
    1. the kptf/pp/orb files are the absolute path, but not check if the files exist
    2. if the INPUT or STRU is not read, the corresponding value will be None
    3. if INPUT defines the stru_file, pseudo_dir or orb_dir, the returned dict will delete the corresponding key in the input_param
    4. extra_files are the files in the folder but not in the list of INPUT, STRU, KPT, PP, ORB
    """
    if not os.path.isdir(ipath):
        print(f"ERROR: {ipath} is not a directory")
        return {
            "input": None,
            "stru": None,
            "kpt": None,
            "pp": [],
            "orb": [],
            "extra_files": []
        }
    
    pwd = os.getcwd()
    os.chdir(ipath)
    
    if os.path.isfile("INPUT"):
        input_param = ReadInput("INPUT")
        struf = input_param.pop("stru_file","STRU")
        kptf = "KPT"
        if "kpt_file" in input_param:
            kptf = input_param["kpt_file"]
            input_param["kpt_file"] = os.path.basename(kptf)
    else:
        print(f"ERROR: INPUT file not found in {ipath}")
        input_param = {}
        kptf = "KPT"
        struf = "STRU"
        
    if not os.path.isfile(struf):
        print(f"ERROR: STRU file '{struf}' not found in {ipath}")
        stru = None
    else:
        stru = AbacusStru.ReadStru(struf)
        if not stru:
            print(f"ERROR: read STRU failed in {ipath}")
            stru = None
            
    pp_path = input_param.pop("pseudo_dir","")
    orb_path = input_param.pop("orb_dir","")
    
    pp_name = []
    orb_name = []
    pp_abs_path = None
    orb_abs_path = None
    if stru:
        pp = stru.get_pp()
        orb = stru.get_orb()
        pp = pp if pp else []
        orb = orb if orb else []
        pp_abs_path = [os.path.abspath(os.path.join(pp_path,i)) for i in pp]
        orb_abs_path = [os.path.abspath(os.path.join(orb_path,i)) for i in orb]
        pp_name = [os.path.basename(i) for i in pp]
        orb_name = [os.path.basename(i) for i in orb]
        if len(pp_name) > 0:
            stru.set_pp(pp_name)
        if len(orb_name) > 0:
            stru.set_orb(orb_name)    
    
    extra_files = []
    for ifile in os.listdir("."):
        if ifile not in ["INPUT",struf,kptf]+pp_name+orb_name and os.path.isfile(os.path.join(ifile)):
            extra_files.append(os.path.abspath(ifile))
    
    kptf = os.path.abspath(kptf) if os.path.isfile(kptf) else None
    os.chdir(pwd)        
    return {
        "input": input_param if input_param else None,
        "stru": stru,
        "kpt": kptf,
        "pp": pp_abs_path,
        "orb": orb_abs_path,
        "extra_files": extra_files
    }

def clean_none_list(*args):
    '''
    For given same length lists, remove the elements that are None in the same position.
    If the length of the lists are different, return the original lists.
    '''
    if len(args) == 0:
        return []
    n = len(args[0])
    for i in args:
        if len(i) != n:
            return args
    new_args = [[] for i in range(len(args))]
    for i in range(n):
        if all([not isinstance(j[i],type(None)) for j in args]):
            for j in range(len(args)):
                new_args[j].append(args[j][i])
    return tuple(new_args)

def run_command(
        cmd,
        shell=True
):
    process = subprocess.Popen(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        shell=shell,
        executable='/bin/bash'
    )
    out = ""
    err = ""
    while True:
        readable, _, _ = select.select(
            [process.stdout, process.stderr], [], [])

        # 读取已经准备好的输出
        for fd in readable:
            if fd == process.stdout:
                line = process.stdout.readline()
                print(line.decode()[:-1])
                out += line.decode()
            elif fd == process.stderr:
                line = process.stderr.readline()
                print("STDERR:", line.decode()[:-1])
                err += line.decode()

        # 如果子进程已经结束，则退出循环
        return_code = process.poll()
        if return_code is not None:
            break
    return return_code, out, err


def cal_cellparam(cell):
    """
    Calculate the parameters of a cell.
    
    Parameters:
        cell (np.ndarray): The cell matrix.
        
    Returns:
        tuple: A tuple containing the lengths and angles of the cell.
    """
    import numpy as np
    a = np.linalg.norm(cell[0])
    b = np.linalg.norm(cell[1])
    c = np.linalg.norm(cell[2])
    
    alpha = np.arccos(np.dot(cell[1], cell[2]) / (b * c)) * 180 / np.pi
    beta = np.arccos(np.dot(cell[0], cell[2]) / (a * c)) * 180 / np.pi
    gamma = np.arccos(np.dot(cell[0], cell[1]) / (a * b)) * 180 / np.pi
    
    return a, b, c, alpha, beta, gamma

def inter_coord(cell1, coord1, 
                cell2, coord2,
                n):
    """
    Interpolate coordinates from one cell to another.
    
    Parameters:
        cell1 (np.ndarray): The first cell matrix.
        coord1 (np.ndarray): Coordinates in the first cell.
        cell2 (np.ndarray): The second cell matrix.
        coord2 (np.ndarray): Coordinates in the second cell.
        n (int): Number of interpolation points.
        
    Returns:
        np.ndarray: Interpolated coordinates in the second cell.
    """
    import numpy as np
    cell1 = np.array(cell1)
    cell2 = np.array(cell2)
    coord1 = np.array(coord1)
    coord2 = np.array(coord2)
    coord1_direct = np.dot(np.linalg.inv(cell1), coord1.T).T
    coord2_direct = np.dot(np.linalg.inv(cell2), coord2.T).T

    coord_diff = coord2_direct - coord1_direct
    # if diff larger than 0.5, then add 1 to the diff
    while np.any(coord_diff > 0.5) or np.any(coord_diff < -0.5):
        coord_diff[coord_diff > 0.5] -= 1
        coord_diff[coord_diff < -0.5] += 1
    coord2_direct = coord1_direct + coord_diff

    maxdiff_idx = np.argmax(np.abs(coord_diff))
    print("Max difference index:", maxdiff_idx)
    print("Max difference value:", np.max(np.abs(coord_diff)))
    
    a1, b1, c1, alpha1, beta1, gamma1 = cal_cellparam(cell1)
    a2, b2, c2, alpha2, beta2, gamma2 = cal_cellparam(cell2)
    print(f"Cell1: a={a1:.2f}, b={b1:.2f}, c={c1:.2f}, alpha={alpha1:.2f}, beta={beta1:.2f}, gamma={gamma1:.2f}")
    print(f"Cell2: a={a2:.2f}, b={b2:.2f}, c={c2:.2f}, alpha={alpha2:.2f}, beta={beta2:.2f}, gamma={gamma2:.2f}")
    
    cells = []
    coords = []
    for i in range(n+1):
        alpha = alpha1 + (alpha2 - alpha1) * i / n
        beta = beta1 + (beta2 - beta1) * i / n
        gamma = gamma1 + (gamma2 - gamma1) * i / n
        
        a = a1 + (a2 - a1) * i / (n - 1)
        b = b1 + (b2 - b1) * i / (n - 1)
        c = c1 + (c2 - c1) * i / (n - 1)
        
        cell = np.array([[a, 0, 0], 
                         [b * np.cos(np.radians(gamma)), b * np.sin(np.radians(gamma)), 0], 
                         [c * np.cos(np.radians(beta)), 
                          c * (np.cos(np.radians(alpha)) - np.cos(np.radians(beta)) * np.cos(np.radians(gamma))) / np.sin(np.radians(gamma)), 
                          c * np.sqrt(1 - np.cos(np.radians(alpha))**2 - np.cos(np.radians(beta))**2 + 2 * np.cos(np.radians(alpha)) * np.cos(np.radians(beta)) * np.cos(np.radians(gamma)))]])

        coord_direct = coord1_direct + coord_diff * i / n
        coord = np.dot(cell, coord_direct.T).T
        
        cells.append(cell)
        coords.append(coord)
    return cells, coords

def interplate_3d_grid(org_data, 
                    new_shape):
    """Interpolate a 3D grid to a new shape using scipy's map_coordinates.
    
    Args:
        org_data (np.ndarray): The original 3D data array.
        new_shape (tuple): The desired shape of the new data array.
        
    Returns:
        np.ndarray: The interpolated 3D data array.
    """
    import numpy as np
    from scipy.ndimage import map_coordinates
    assert len(org_data.shape) == len(new_shape) == 3, "org_data and new_shape should be 3D"
    
    old_nx, old_ny, old_nz = org_data.shape
    new_nx, new_ny, new_nz = new_shape
    x = np.linspace(0, old_nx, new_nx, endpoint=False)
    y = np.linspace(0, old_ny, new_ny, endpoint=False)
    z = np.linspace(0, old_nz, new_nz, endpoint=False)
    X, Y, Z = np.meshgrid(x, y, z, indexing='ij')
    coords = np.array([X.flatten(), Y.flatten(), Z.flatten()])
    
    padding = 3
    data = np.pad(org_data, ((padding,padding),(padding,padding),(padding,padding)), mode="wrap")
    coords += padding
    
    return map_coordinates(data, coords, order=2, mode='wrap').reshape(new_shape)


def read_gaussian_cube(fcube: str):
    """read the Gaussian cube format volumetric data.
    The detailed format information can be found here:
    https://paulbourke.net/dataformats/cube/

    In brief, 

    ```
    [comment line1]
    [comment line2]
    [natom] [x-origin] [y-origin] [z-origin]
    [nx] [e11 in a.u.] [e12 in a.u.] [e13 in a.u.]
    [ny] [e21 in a.u.] [e22 in a.u.] [e23 in a.u.]
    [nz] [e31 in a.u.] [e32 in a.u.] [e33 in a.u.]
    [Z1] [chg1] [x1 in a.u.] [y1 in a.u.] [z1 in a.u.]
    [Z2] [chg2] [x2 in a.u.] [y2 in a.u.] [z2 in a.u.]
    ...
    [Znatom] [xnatom in a.u.] [ynatom in a.u.] [znatom in a.u.]
    [data1] [data2] [data3] ... [data6]
    [data7] [data8] [data9] ... [data12]
    ...
    ```
    
    Args:
        fcube (str): the file name of the cube file.
    
    Returns:
        data (dict): a dictionary containing the following
    """
    import numpy as np
    with open(fcube, 'r') as f:
        lines = f.readlines()
    
    out = {}
    out["comment 1"] = lines[0].strip()
    out["comment 2"] = lines[1].strip()

    natom, x_origin, y_origin, z_origin = map(float, lines[2].split())
    out["natom"] = natom
    out["origin"] = (x_origin, y_origin, z_origin)

    out["nx"] = int(lines[3].split()[0])
    out["ny"] = int(lines[4].split()[0])
    out["nz"] = int(lines[5].split()[0])
    e11, e12, e13 = map(float, lines[3].split()[1:4])
    e21, e22, e23 = map(float, lines[4].split()[1:4])
    e31, e32, e33 = map(float, lines[5].split()[1:4])
    out["R"] = np.array([[e11, e12, e13], [e21, e22, e23], [e31, e32, e33]])

    atomz, chg, coords = [], [], []
    for line in lines[6:6+int(natom)]:
        atomz.append(int(line.split()[0]))
        chg.append(float(line.split()[1]))
        coords.append(list(map(float, line.split()[2:])))
    out["atomz"] = atomz
    out["chg"] = chg
    out["coords"] = coords

    data = []
    for line in lines[6+int(natom):]:
        data.extend(list(map(float, line.split())))
    #out["data"] = np.array(data).reshape(int(nx), int(ny), int(nz)) # is it necessary to reshape?
    out["data"] = np.array(data) # not reshaped
    return out

def write_gaussian_cube(data: dict, fcube: str, ndigits: int = 6):
    """write the Gaussian cube format volumetric data.
    Data should be organized as what is returned by read_cube().
    
    (if present) the ndigits will be used to control the number of digits of volumetric data,
    while the coordination data in header of cube is fixed to 6 digits.
    
    format of each part is defined in the following with format string:
    ```
    %s
     %s
     %4d %11.6f %11.6f %11.6f
     %4d %11.6f %11.6f %11.6f
     %4d %11.6f %11.6f %11.6f
     %4d %11.6f %11.6f %11.6f
     %4d %11.6f %11.6f %11.6f %11.6f
    ...
     %[width].[ndigits]f %[width].[ndigits]f %[width].[ndigits]f %[width].[ndigits]f %[width].[ndigits]f %[width].[ndigits]f
    ...
    ```

    Args:
        data (dict): the dictionary containing the cube data.
        fcube (str): the file name of the cube file.
        ndigits (int): the number of digits to be used for the volumetric data.
                        Default is 6, which is the same as the original cube format.
    
    Returns:
        None
    """
    
    width = ndigits + 7
    # temporarily there is no more format controlling options.

    with open(fcube, 'w') as f:
        f.write(data["comment 1"] + "\n")
        f.write(data["comment 2"] + "\n")
        f.write(" %4d %11.6f %11.6f %11.6f\n" % (data["natom"], data["origin"][0], data["origin"][1], data["origin"][2]))
        f.write(" %4d %11.6f %11.6f %11.6f\n" % (data["nx"], data["R"][0][0], data["R"][0][1], data["R"][0][2]))
        f.write(" %4d %11.6f %11.6f %11.6f\n" % (data["ny"], data["R"][1][0], data["R"][1][1], data["R"][1][2]))
        f.write(" %4d %11.6f %11.6f %11.6f\n" % (data["nz"], data["R"][2][0], data["R"][2][1], data["R"][2][2]))
        for i in range(int(data["natom"])):
            f.write(" %4d %11.6f %11.6f %11.6f %11.6f\n" % (data["atomz"][i], data["chg"][i], data["coords"][i][0], data["coords"][i][1], data["coords"][i][2]))
        for i in range(0, len(data["data"]), 6):
            f.write(" ".join([f"%{width}.{ndigits}e" % x for x in data["data"][i:i+6]]))
            f.write("\n")
    return

def read_abacus_chg(fcube: str):
    """Read the ABACUS CHG cube format volumetric data.
    
    Args:
        fcube (str): the file name of the CHG cube file.
    
    Returns:
        data (dict): a dictionary containing the following
            - "data": np.ndarray, the volumetric data array. unit in e/A^3
            - "cell": np.ndarray, the cell matrix in Angstrom
            - "coords": np.ndarray, the atomic coordinates in Angstrom
            - "elements": list, the atomic numbers
    """
    from abacustest.constant import BOHR2A
    data = read_gaussian_cube(fcube)
    import numpy as np
    out = {}
    out["data"] = np.array(data["data"]).reshape((data["nx"], data["ny"], data["nz"]))
    out["data"] *= 1.0 / BOHR2A**3  # convert from e/Bohr^3 to e/A^3
    out["cell"] = np.array(data["R"]) * np.array([[data["nx"]], [data["ny"]], [data["nz"]]]) * BOHR2A  # in Angstrom
    out["coords"] = np.array(data["coords"]) * BOHR2A  # in Angstrom
    out["elements"] = data["atomz"]
    return out

def write_abacus_chg(data,
                     cell,
                     coords,
                     elements, 
                     fcube: str, 
                     ndigits: int = 6):
    """Write the ABACUS CHG cube format volumetric data.
    
    Args:
        data (np.ndarray): the volumetric data array. unit in e/A^3
        cell (np.ndarray): the cell matrix in Angstrom
        coords (np.ndarray): the atomic coordinates in Angstrom
        elements (list): the atomic numbers
        fcube (str): the file name of the CHG cube file.
        ndigits (int): the number of digits to be used for the volumetric data.
                        Default is 6, which is the same as the original cube format.
    """
    from abacustest.constant import BOHR2A
    import numpy as np
    data = np.array(data)
    cell = np.array(cell)
    coords = np.array(coords)
    assert len(data.shape) == 3, "data should be a 3D array"
    assert cell.shape == (3, 3), "cell should be a 3x3 array"
    assert coords.shape[1] == 3, "coords should be a Nx3 array"
    assert len(elements) == coords.shape[0], "elements should have the same length as coords"
    
    out = {}
    out["comment 1"] = "CHG cube file generated by abacustest"
    out["comment 2"] = "unit: e/Bohr^3"
    out["natom"] = len(elements)
    out["origin"] = (0.0, 0.0, 0.0)  # in Bohr
    out["nx"], out["ny"], out["nz"] = data.shape
    out["R"] = cell / np.array([[out["nx"]], [out["ny"]], [out["nz"]]]) / BOHR2A  # in Bohr
    out["atomz"] = elements
    out["chg"] = [0.0] * len(elements)
    out["coords"] = coords / BOHR2A  # in Bohr
    out["data"] = (data * BOHR2A**3).flatten()  # convert from e/A^3 to e/Bohr^3 and flatten
    write_gaussian_cube(out, fcube, ndigits)
    return

def read_abacus_pot(fcube: str):
    """Read the ABACUS POT cube format volumetric data.
    
    Args:
        fcube (str): the file name of the POT cube file.
    
    Returns:
        data (dict): a dictionary containing the following
            - "data": np.ndarray, the volumetric data array. unit in eV
            - "cell": np.ndarray, the cell matrix in Angstrom
            - "coords": np.ndarray, the atomic coordinates in Angstrom
            - "elements": list, the atomic numbers
    """
    from abacustest.constant import BOHR2A, RY2EV
    data = read_gaussian_cube(fcube)
    import numpy as np
    out = {}
    out["data"] = np.array(data["data"]).reshape((data["nx"], data["ny"], data["nz"]))
    out["data"] *= -1 * RY2EV  # convert from Ry to eV and change the sign
    out["cell"] = np.array(data["R"]) * np.array([[data["nx"]], [data["ny"]], [data["nz"]]]) * BOHR2A  # in Angstrom
    out["coords"] = np.array(data["coords"]) * BOHR2A  # in Angstrom
    out["elements"] = data["atomz"]
    return out

def check_abacus_inputs(job: str) -> Tuple[bool, str]:
    """
    Check if the ABACUS input files in the given job directory are valid.
    
    Parameters:
        job (str): The job directory.
        
    Returns:
        Tuple[bool, str]: A tuple containing a boolean indicating if the inputs are valid,
                          and a string message.
    """
    if not os.path.isdir(job):
        return False, f"{job} is not a directory"
    
    # check INPUT file
    if not os.path.isfile(os.path.join(job, "INPUT")):
        return False, f"INPUT file not found in {job}"
    
    input_param = ReadInput(os.path.join(job, "INPUT"))
    struf = input_param.get("stru_file", "STRU")
    # check STRU file
    if not os.path.isfile(os.path.join(job, struf)):
        return False, f"STRU file '{struf}' not found in {job}"
    
    stru = AbacusStru.ReadStru(os.path.join(job, struf))
    if not stru:
        return False, f"Read STRU failed in {job}"
    
    # check K Point file
    kspacing = input_param.get("kspacing", None)
    basis = input_param.get("basis_type", "pw")
    gamma_only = input_param.get("gamma_only", False)
    kptf = input_param.get("kpoint_file", "KPT")
    if kspacing is None and (basis == "pw" or (basis == "lcao" and not gamma_only)) and not os.path.isfile(os.path.join(job, kptf)):
        return False, f"KPT file '{kptf}' not found in {job} and kspacing is not set."
    
    # check pp/orbital files
    pp_path = input_param.get("pseudo_dir", "")
    orb_path = input_param.get("orbital_dir", "")
    pp = stru.get_pp()
    orb = stru.get_orb()
    if pp is None:
        return False, f"Pseudopotential files not defined in STRU file '{struf}' in {job}"
    
    # check pp_file
    pp = [os.path.join(job, pp_path, i) for i in pp]
    for ipp in pp:
        if not os.path.isfile(ipp):
            return False, f"Pseudopotential file '{ipp}' not found in {job}"
    
    # check orbital files
    if basis == "lcao" or input_param.get("onsite_radius", None):
        if orb is None:
            return False, f"Orbital files not defined in STRU file '{struf}' in {job}"
        orb = [os.path.join(job, orb_path, i) for i in orb]
        for iorb in orb:
            if not os.path.isfile(iorb):
                return False, f"Orbital file '{iorb}' not found in {job}"
    
    # check orther INPUT parameters
    ks_solver = input_param.get("ks_solver", None)
    device = input_param.get("device", "cpu")
    if basis == "pw" and ks_solver is not None and ks_solver not in ["cg", "dav", "dav_subspace", "bpcg"]:
        return False, f"Invalid ks_solver '{ks_solver}' in INPUT file in {job}. It should be 'cg', 'dav', 'bpcg', or 'dav_subspace' for PW basis."
    if basis == "lcao" and ks_solver is not None:
        if device == "cpu" and ks_solver not in ["genelpa", "scalapack_gvx","lapack", "elpa"]:
            return False, f"Invalid ks_solver '{ks_solver}' in INPUT file in {job}. It should be 'genelpa', 'scalapack_gvx', 'elpa', or 'lapack' for LCAO basis with CPU device."
        if device == "gpu" and ks_solver not in ["cusolver", "cusolvermp", "elpa"]:
            return False, f"Invalid ks_solver '{ks_solver}' in INPUT file in {job}. It should be 'cusolver', 'cusolvermp', or 'elpa' for LCAO basis with GPU device."
    
    # check nspin and soc
    nspin = input_param.get("nspin", None)
    soc = IsTrue(input_param.get("lspinorb", None))
    noncollinear = IsTrue(input_param.get("noncolin", None))
    if nspin is not None:
        if nspin not in [1, 2, 4, "1", "2", "4"]:
            return False, f"Invalid 'nspin' value: {nspin}. It should be 1 (no spin), 2 (spin polarized), or 4 (non-collinear spin)."
        nspin = int(nspin)
        if nspin in [1, 2] and (soc or noncollinear):
            return False, "Spin-orbit coupling ('lspinorb') or non-collinear spin ('noncolin') is set, but nspin is not 4. Please set nspin to 4 for non-collinear spin calculations."
        if nspin == 4 and noncollinear == False:
            return False, "Non-collinear spin ('noncolin') is set to False, but nspin is 4. Please set noncollinear to True for non-collinear spin calculations."
    elif soc and noncollinear == False:
        return False, "Spin-orbit coupling ('lspinorb') is set, but noncollinear ('noncolin') is set to False. Please set noncollinear to True for non-collinear spin calculations."
    
    # check out_mul
    if basis == "pw" and input_param.get("out_mul", None):
        return False, "'out_mul' is invalid for PW basis. Please remove it from INPUT file."
    
    if basis == "pw" and nspin == 2 and input_param.get("onsite_radius", None):
        return False, "'onsite_radius' is invalid for PW basis with nspin=2. Please remove it from INPUT file."
    
    return True, "All input files are valid"

def reciprocal_lattice(cell):
    """Calculate the reciprocal lattice vectors from the real space cell vectors."""
    import numpy as np
    volume = np.dot(cell[0], np.cross(cell[1], cell[2]))
    b1 = 2 * np.pi * np.cross(cell[1], cell[2]) / volume
    b2 = 2 * np.pi * np.cross(cell[2], cell[0]) / volume
    b3 = 2 * np.pi * np.cross(cell[0], cell[1]) / volume
    return np.array([b1, b2, b3])

def chg2pot(chg, cell):
    """Poisson equation solver using FFT method.
    
    Args:
        chg (np.ndarray): The charge density in e/Angstrom^3 with shape (nx, ny, nz).
        cell (np.ndarray): The cell vectors in Angstrom.
    
    Returns:
        potential (np.ndarray): The electrostatic potential in V.
    """
    # chg in unit: e/Angstrom^3
    import numpy as np
    from ase import units
    
    nx, ny, nz = chg.shape
    
    cell_rec = reciprocal_lattice(cell)

    kx = np.fft.fftfreq(nx, d= 1 / nx)
    ky = np.fft.fftfreq(ny, d= 1 / ny)
    kz = np.fft.fftfreq(nz, d= 1 / nz)
    Kx, Ky, Kz = np.meshgrid(kx, ky, kz, indexing='ij')
    
    # expand kx, ky, kz, to 1D array
    g = np.vstack((Kx.flatten(), Ky.flatten(), Kz.flatten())).T  

    K = np.dot(g, cell_rec)
    K_squared = np.sum(K**2, axis=1).reshape((nx, ny, nz))

    rho_fft = np.fft.fftn(chg) # in unit e/Angstrom^3
    
    phi_fft = np.zeros_like(rho_fft)
    K_squared[0,0,0] = 1  # avoid division by zero
    phi_fft = rho_fft/ K_squared # unit: e/Angstrom
    

    eps0 = units._eps0 
    # eps0 = 8.85418781762039e-12 F/m = 8.85418781762039e-12 C/(Vm) = 8.85418781762039e-12 *  1e/-1.602e-19 / 1e10 e/(V Angstrom)
    # = 0.005526 e/(V Angstrom)
    
    potential = np.fft.ifftn(phi_fft).real / eps0 * units._e * -1 * 1e10 # unit: V
    # p = rho / (eps0 * k^2) = e/A / (F/m) = e * m / (A * F) = e * 1e10 / F = -1 * units._e C * 1e10 V / C
    # = -1 * units._e * 1e10 V
    
    return potential

def data2dpdata(lattice, coords, atom_types):
    """Convert lattice, coordinates, and atom types to a dpdata System object.
    
    Args:
        lattice (list or np.ndarray): The lattice vectors, unit in Angstrom.
        coords (list or np.ndarray): The atomic cartesian coordinates, unit in Angstrom.
        atom_types (list): The atomic types as atomic numbers.
        
    Returns:
        dpdata.System: The dpdata System object.
    """
    import dpdata
    from abacustest.constant import PERIOD_DICT_ELEMENT
    import numpy as np
    
    elements = [PERIOD_DICT_ELEMENT.get(num, num) for num in atom_types]
    element_uniq = []
    for i in elements:
        if i not in element_uniq:
            element_uniq.append(i)
    
    atom_types = [element_uniq.index(i) for i in elements]
    atom_numbers = [elements.count(i) for i in element_uniq]
    return dpdata.System(data={
        "atom_numbs": atom_numbers,
        "atom_names": element_uniq,
        "atom_types": np.array(atom_types),
        "orig": np.array([0, 0, 0]),
        "cells": np.array([lattice]),
        "coords": np.array([coords])
    })


import os,glob,json,traceback,sys
from abacustest.lib_collectdata.collectdata import RESULT
from abacustest.collectdata import parse_value

class Metrics:
    def __init__(self,dft_type="abacus",metrics_name=[],newmethods=[],path=["."],modules=[]):
        self.dft_type = dft_type
        self.metrics_name = metrics_name
        self.newmethods = newmethods
        self.path = path
        self.modules = modules
        pass
    
    def get_metrics(self,save_file = None):
        allvalue = {}
        print("metrics setting (getcwd(), path=, newmethods=, dft_type=, modules=):",os.getcwd(),self.path,self.newmethods,self.dft_type,self.modules)
        print("os.listdir:",os.listdir("."))
        for ipath in self.path:
            for iipath in glob.glob(ipath):
                if not os.path.isdir(iipath):
                    continue
                try:
                    print("get metrics from:",iipath)
                    result = RESULT(fmt=self.dft_type,newmethods=self.newmethods,path=iipath,modules=self.modules)
                    if len(self.metrics_name) == 0:
                        self.metrics_name = result.AllMethod().keys()
                    allvalue[iipath] = parse_value(result,self.metrics_name) 
                    
                    #write version.dat, this file is used by report section
                    if not os.path.isfile("version.dat"):
                        version = result["version"]
                        version = "unknow" if version == None else version
                        with open("version.dat",'w') as f:
                            f.write(version)
                except:
                    traceback.print_exc()
        if save_file != None:
            json.dump(allvalue,open(save_file,'w'),sort_keys=True,indent=4) 
        return allvalue
    
    @staticmethod
    def TransferMetricsOPIO(metrics_io):
        if isinstance(metrics_io,dict):
            poin_metrics = [metrics_io]
        elif isinstance(metrics_io,list):
            poin_metrics = metrics_io
        else:
            poin_metrics = []
        return poin_metrics 

    @staticmethod
    def ParseMetricsOPIO(metrics_io,example_path):
        all_dfttype = {0:"abacus",1:"qw",2:"vasp"}
        dft_type = metrics_io.get("dft_type",None)
        dftt = None
        if dft_type == None:
            return None
        elif isinstance(dft_type,int):
            if dft_type not in all_dfttype:
                print("Unknow dft type '%d'. Supportted:" % dft_type,str(all_dfttype))
            else:
                dftt = all_dfttype[dft_type]
        elif isinstance(dft_type,str):
            if dft_type.lower() == "abacus":
                dftt = "abacus"
            elif dft_type.lower() == "qe":
                dftt = "qe"
            elif dft_type.lower() == "vasp":
                dftt = "vasp"
            else:
                print("Unknow dft type '%s'. Supportted:" % dft_type,str(all_dfttype))
        else:
            print("Unknow dft type '",dft_type,"'. Supportted:",str(all_dfttype))

        if "path" not in metrics_io:
            epath = example_path
        else:
            epath = metrics_io["path"]
            
        if dftt == None:
            return None
        else:
            print(metrics_io)
            return Metrics(dft_type=dftt,
                           path= epath,
                       metrics_name= metrics_io.get("metrics_name",[]),
                       newmethods=metrics_io.get("newmethods",[]),
                       modules = metrics_io.get("modules",[]))       

    @staticmethod
    def SuperMetricsResult(super_metrics_setting):
        if not super_metrics_setting.get("result_file"):
            return None,None,None
        from abacustest import outresult
        allresults = outresult.GetAllResults(super_metrics_setting)
        if not allresults:
            return None,None,None
        split_example=None if len(allresults["type_name"]) == 1 else "----"
        cc_outparam,allparam_value = outresult.OutParam(allresults,split_example=split_example)
        cc_outmetrics,allmetric_value = outresult.OutMetrics(allresults,allparam_value)
        report = ""
        if len(allresults["metrics"]) > 0:
            report += cc_outmetrics
            save_file = super_metrics_setting.get("save_file","superMetric.json")
            json.dump(allmetric_value,open(save_file,'w'),indent=4)
        report += cc_outparam

        type_name = allresults["type_name"]
        if allmetric_value and len(type_name) > 1:
            #if there has more than 1 type, transfer allmetric_value to one type one value
            allmetric_value_final = {}
            for k,v in allmetric_value.items():
                for i,itype in enumerate(type_name):
                    allmetric_value_final[k+"_" +itype] = v[i]
        else:
            allmetric_value_final = allmetric_value     

        return allparam_value,allmetric_value_final,report

    
    @staticmethod
    def rotate_metrics(dict1):
        '''
        dict1 = {example1: {metric1:,metric2:,...}, example2: {}}
        will be rotated to:
        {metric1:[example1_value,example2_value],metric2:[example1_value,example2_value,...]}
        '''
        allkey = [i for i in dict1.keys()]
        allmetrics = dict1[allkey[0]].keys()
        newdict = {}
        for imetric in allmetrics:
            newdict[imetric] = []
            for iexample in allkey:
                newdict[imetric].append(dict1.get(iexample,{}).get(imetric,None))
        return allkey,newdict
    
    @staticmethod
    def Transfer2Table(dict1):
        '''
        dict1 = {example1: {metric1:,metric2:,...}, example2: {}}
        will be transfered to:
        [{"sample":example1,"metric1":,"metric2":,...},{}]
        '''
        from dp.tracking import Table
        allkey = [i for i in dict1.keys()]
        allmetrics = dict1[allkey[0]].keys()
        newlist = []
        for ikey in allkey:
            newlist.append({"sample":ikey})
            for imetric in allmetrics:
                newlist[-1][imetric] = dict1[ikey].get(imetric,None)
        return Table(newlist)
    
    
def ReadMetrics(poin_metrics,do_upload_tracking,example_path):
    """
    poin_metrics is a list of dict
    poin_metrics = [
        {
            "value_from_file": str,     #a json file that has the metrics and value, which will also be uploaded to tracking, should be {"example":{"key":value,...},...} 
            "group_name": str,  #self-defined group name, only used in context for tracking
            "path": ["*"],      #the path of dft jobs
            "dft_type": "abacus",   #dft type, can be abacus/qe/vasp
            "metrics_name": [],     #the name of metrics that will collected from dft job
            "newmethods": [str],    # self-defined methods used in abacustest collectdata
            "modules": [str],       # the modules that abacustest collectdata will load
            "save_file": "metrics.json",  #file to store the value of metrics
        },
        {"path": ["*"], ...}, ...
    ]

    return tracking_values
    tracking_values = [(metrics_value,context),...]
    metrics_value = {metric_name:value}
    
    """
    tracking_values = []
    for im,imetric in enumerate(poin_metrics):
        metrics_value = {}

        metric_form_calc = Metrics.ParseMetricsOPIO(imetric,example_path)
        if metric_form_calc:
            value = metric_form_calc.get_metrics(save_file=imetric.get("save_file","metrics.json"))
        else:
            value = {}

        if do_upload_tracking:
            metric_from_file = imetric.get("value_from_file",None)
            if metric_from_file: 
                if os.path.isfile(metric_from_file):
                    try:
                        for k,v in json.load(open(metric_from_file)).items():
                            if k not in value:
                                value[k] = {}
                            for ik,iv in v.items():
                                value[k][ik] = iv
                    except:
                        traceback.print_exc()
                else:
                    print("Can not find file %s" % metric_from_file,file=sys.stderr)
                    print("Current path: %s, listdir:" % os.path.abspath("."),os.path.listdir("."),file=sys.stderr)
                    
            if value: 
                try:
                    examples,new_dict = Metrics.rotate_metrics(value)
                    for k,v in new_dict.items():
                        metrics_value[k] = v
                    metrics_value["sample_name_%d" % im] = examples
                    metrics_value["metrics_%d"%im] = Metrics.Transfer2Table(value) 
                except:
                    traceback.print_exc()
                    
            context = {"subset":"metrics%d"%im,"datatype":"metrics"}
            if "group_name" in imetric:
                context["group_name"] = imetric.get("group_name")

            if metrics_value:
                tracking_values.append((metrics_value,context))

    return tracking_values

def ReadSuperMetrics(poin_super_metrics,do_upload_tracking):
    """
    poin_super_metrics = [{
        "group_name": str,  # a json file that has the metrics and value, which will also be uploaded to tracking
        "value_from_file": str, # self-defined super metrics, can also define some files
                                # {key:{"type":"image","file":"file.png"}}
        "save_file": "superMetric.json",    #same as that defined in outresult report
        "result_file": ["metrics.json"],
        "example_name_idx":0,
        "outparams":[
            ["ScfSteps", ["scf_steps"], 0],...
        ],
        "metrics":[{}]
    },...]

    return tracking_summary
    tracking_summary = [(metrics_value,context),...]
    metrics_value = {metric_name:value}

    """
    tracking_summary = []  
    log = ""             
    for isuper,super_metrics_setting in enumerate(poin_super_metrics):
        super_metric_value = {}
        allparam_value,allmetric_value,report = Metrics.SuperMetricsResult(super_metrics_setting)

        if do_upload_tracking:
            if allmetric_value:
                for k,v in allmetric_value.items():
                    super_metric_value[k] = v
                from dp.tracking import Table
                super_metric_value["super_metrics_%d"%isuper] = Table([allmetric_value])
            if report:
                super_metric_value["report"] = report
                log += report

            super_metric_from_file = super_metrics_setting.get("value_from_file",None)
            if super_metric_from_file:
                if os.path.isfile(super_metric_from_file):
                    try:
                        for k,v in json.load(open(super_metric_from_file)).items():
                            super_metric_value[k] = v
                    except:
                        traceback.print_exc()
                else:
                    print("Can not find file %s" % super_metrics_setting.get("value_from_file"),file=sys.stderr)
                    print("Current path: %s, listdir:" % os.path.abspath("."),os.path.listdir("."),file=sys.stderr)
            if super_metric_value:
                context = {"subset":"super_metrics%d"%isuper,"datatype":"super_metrics"}
                if "group_name" in super_metrics_setting:
                    context["group_name"] = super_metrics_setting.get("group_name")
                tracking_summary.append((super_metric_value,context))
    
    return tracking_summary,log


from dp.launching.typing.basic import BaseModel, Int, String, Float, List, Optional, Union, Dict
from dp.launching.typing import (
    BaseModel,
    Set,
    Boolean,
    Field,
    DflowAccessToken,
    DflowArgoAPIServer,
    DflowK8sAPIServer,
    DflowStorageEndpoint,
    DflowStorageRepository,
    BohriumMachineType,
    BohriumImage,
    BohriumPlatform,
    BohriumJobType,
    BohriumUsername,
    BohriumPassword,
    BohriumProjectId,
    BenchmarkLabels,
    BenchmarkTags
)
from enum import Enum
from typing import Literal
import re

from . import comm_class


class Image(BaseModel):
    type: Literal["Local Image"]
    image: String = Field(default="registry.dp.tech/dptech/abacustest:latest",
                          title="",
                          description="",)


class ImageBohrium(BaseModel):
    type: Literal["Use Bohrium Image"]

    image: String = Field(default=BohriumMachineType("registry.dp.tech/dptech/abacustest:latest"),
                          title="Bohrium Image Address",
                          description="",)
    bohrium_machine_type: BohriumMachineType = Field(
        default=BohriumMachineType("c2_m4_cpu"))
    bohrium_job_type: BohriumJobType = Field(default=BohriumJobType.CONTAINER)
    bohrium_plat_form: BohriumPlatform = Field(default=BohriumPlatform.ALI)
    on_demand: Boolean = Field(default=False)
    
    #bohrium_machine_type: String = Field(default="c2_m4_cpu")
    #bohrium_job_type: String = Field(default="container")
    #bohrium_plat_form: String = Field(default="ali")



class PredftImageSet(BaseModel):
    predft_image_set: Union[Image,
                     ImageBohrium,
                     comm_class.ImageDispatcher,] = Field(discriminator="type",
                                           description="IMAGE and MACHINE used in PRE DFT")

class PredftGroupSizeSet(BaseModel):
    predft_group_size: Int = Field(
        default=0, description="Number of examples running in each parallel machine. If not set, all examples will be run by serial in one machine.", ge=0)


class PredftCommandSet(BaseModel):
    predft_command: String = Field(default="",
                                   description="Command of predft in each example. Please note that the program will first enter each folder before executing this command.\
If executing the command will generate some new example directories, please write these directories to a file named as 'example.txt', and each directory one line.",)
    
    #do not support self-defined filename in launching, and use default filename "example.txt"
    #predft_work_directories_filename: String = Field(default="",
    #                                               description="If executing the predft_command will generate some new example directories, please write these directories to a file with one line per directory. And enter the file name here",)

def parse_image_set(image_set):
    if isinstance(image_set, ImageBohrium):
        return {
            "image": image_set.image,
            "bohrium": {
                "scass_type": image_set.bohrium_machine_type,
                "job_type": image_set.bohrium_job_type,
                "platform": image_set.bohrium_plat_form,
                "on_demand": 1 if image_set.on_demand else 0
            }
        }
    elif isinstance(image_set, Image):
        return {
            "image": image_set.image
        }
    elif isinstance(image_set, comm_class.ImageDispatcher):
        return comm_class.ImageDispatcher.construct_dispatcher(image_set)
    else:
        raise ValueError(f"Unknown image set type: {type(image_set)}")

def construct_input(datas,opts,logs):
    # datas is a dict of examples, created by comm_class_exampleSource.read_source
    #read predft
    need_predft = False
    logs.iprint("read pre dft setting ...")
    pre_dft = {}
    
    #read predft example
    if datas.get("predft_example"):
        pre_dft["example"] = datas.get("predft_example")    
        logs.iprint("\texample:",pre_dft["example"])
    
    #read predft extra files
    if datas.get("predft_extrafile"):
        pre_dft["extra_files"] = datas.get("predft_extrafile")
        logs.iprint("\rpredft_extrafile:",pre_dft["extra_files"])
        
    #read predft command
    if hasattr(opts,"predft_command") and opts.predft_command.strip() != "":
        need_predft = True
        logs.iprint("\tcommand:",opts.predft_command)
        pre_dft["command"] = opts.predft_command
        #pre_dft["work_directories_filename"] = opts.predft_work_directories_filename
    
    #read predft ngroup        
    #if hasattr(opts,"predft_group_size") and opts.predft_group_size > 0:
    #    logs.iprint("\tgroup size:",opts.predft_group_size)
    #    pre_dft["group_size"] = opts.predft_group_size
    
    #read predft image
    if need_predft and hasattr(opts,"predft_image_set"):
        image_set = parse_image_set(opts.predft_image_set)   
        logs.iprint("\timage:",image_set.get("image",""))
        for k,v in image_set.items():
            pre_dft[k] = v
        if "bohrium" in pre_dft:   
            logs.iprint("\tbohrium:",pre_dft["bohrium"])
        elif "dispatcher" in pre_dft:
            logs.iprint("\tdispatcher:",pre_dft["dispatcher"]["resources_dict"])
            
    return need_predft,pre_dft


import argparse,importlib,glob,os
from abacustest.lib_model.model import Model

def ReadAllModels():
    return {imodel.model_name():imodel for imodel in Model.__subclasses__()} 

def ModelArgs(parser):  
    allmodels = [os.path.basename(i)[:-3] for i in glob.glob(os.path.join(os.path.dirname(__file__),"lib_model/model_*.py"))]
    for module in allmodels:
        importlib.import_module(f"abacustest.lib_model.{module}")   
    subparser = parser.add_subparsers(dest="model")
    
    models = sorted(Model.__subclasses__(), key=lambda x: x.model_name().lower())
    
    for imodel in models:
        model_name = imodel.model_name()
        iparser = subparser.add_parser(model_name, help=imodel.description())
        iparser.description = imodel.description()
        subsubparser = iparser.add_subparsers(dest="modelcommand")
        iparser_prepare = subsubparser.add_parser("prepare", help="Prepare the model")
        iparser_postprocess = subsubparser.add_parser("post", help = "Post-process the model")
        imodel.prepare_args(iparser_prepare)
        imodel.postprocess_args(iparser_postprocess)
        imodel.add_args(iparser)
    parser.description = "Prepare and post-process the specified model"

    return parser

def RunModel(param):
    allmodels = ReadAllModels()
    imodel = allmodels[param.model]()
    
    print(f"Model: {param.model}")
    if param.modelcommand == "prepare":
        imodel.run_prepare(param)
    elif param.modelcommand == "post":
        imodel.run_postprocess(param)
    else:
        imodel.run(param)

def main():
    parser = argparse.ArgumentParser()
    param = ModelArgs(parser).parse_args()
    RunModel(param)
    
if __name__ == "__main__":
    main()


import os,sys,json
from .result import Result
from . import comm

class ResultQe(Result):
    _PARAM_DIC = {}

    def __init__(self,path = ".",output = None,resultREF="resultREF.json"):
        super().__init__()
        self.PATH = path   #the path of QE job

        self.XMLf = os.path.join(self.PATH,"pwscf.xml")
        self.OUTPUTf = output if output != None else comm.FindOutput(self.PATH,keyinfo='Program PWSCF')

        self.OUTPUT = comm.ReadFile(self.OUTPUTf,warn=False)
        self.XMLROOT = comm.ReadXmlFile(self.XMLf,warn=True)

        self.resultREF = {}
        if resultREF and os.path.isfile(resultREF):
            self.resultREF = json.load(open(resultREF))



from ..resultVasp import ResultVasp
from ..comm_funcs.bda import BasicPropertyVasp
import os

class BDA(ResultVasp):
    @ResultVasp.register( bda_mag_moment="mag_moment of some metal element",
                          bda_bond_length="bond_length of some metal element")
    def GetBdaInfo(self):
        cwd = os.getcwd()
        os.chdir(self.PATH)
        output = BasicPropertyVasp('OUTCAR', 'CONTCAR')
        self["bda_mag_moment"] = output.magnetic_moment('TM')
        self["bda_bond_length"] = output.bond_length()
        os.chdir(cwd)
